//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
//#pragma resource "*.dfm"
//---------------------------------------------------------------------------
#include "TDirectDraw.h"
//---------------------------------------------------------------------------
static inline TDirectDraw *ValidCtrCheck()
{
	return new TDirectDraw(NULL);
}
//---------------------------------------------------------------------------
namespace Tdirectdraw
{
	void __fastcall Register()
	{
		TComponentClass classes[1] = {__classid(TDirectDraw)};
		RegisterComponents("DirectDraw", classes, 0);
	}
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::TDirectDraw(TComponent* Owner)
	: TComponent(Owner)
{
// create things required for debugging
fErrorValue = 0;
fErrorCount = 0;
fErrorStrings = new TStringList();
fErrorStrings->Sorted = false;
fErrorStrings->Duplicates = dupAccept;
fDebugging = true;

Log(LOG_IN,"TDirectDraw::Constructor()","Owner = >>> " + Owner->Name + " <<<");

// initialize some default values
Log(LOG_INFO,"TDirectDraw::Constructor()","about to initialize internals");
fDirectDraw = NULL;
fPrimarySurface = NULL;
fBackBuffer = NULL;

fWidth = 800;
fHeight = 600;
fBitsPerPixel = 16;
fBackBufferCount = 1;
fHandle = 0;

fDisplayModeChanged = false;
fExclusiveModeSet = false;
fBackBufferLocked = false;
fPrimarySurfaceLocked = false;
fBackBufferUsedLast = false;
fPrimarySurfaceUsedLast = false;
fEnabled = false;

fBackBufferDCRetrieved = false;
fPrimarySurfaceDCRetrieved = false;

// create our DirectDraw object, or skip out if it doesn't work
Log(LOG_INFO,"TDirectDraw::Constructor()","about to create the internal IDirectDraw object");
fErrorValue = DirectDrawCreate( NULL, &fDirectDraw, NULL );
if (fErrorValue != DD_OK)
	{
    Error(fErrorValue,"TDirectDraw::Constructor()", "Could not create the internal IDirectDraw object");
    return;
    }
Log(LOG_NOTE,"TDirectDraw::Constructor()","the internal IDirectDraw object has been created successfully");

// setup the surface description we use for locking / unlocking surfaces
Log(LOG_NOTE,"TDirectDraw::Constructor()","setting up the locking / unlocking surface descriptions");
memset(&fPrimarySurfaceLockingDesc,0,sizeof(fPrimarySurfaceLockingDesc));
fPrimarySurfaceLockingDesc.dwSize = sizeof(fPrimarySurfaceLockingDesc);
memset(&fBackBufferLockingDesc,0,sizeof(fBackBufferLockingDesc));
fBackBufferLockingDesc.dwSize = sizeof(fBackBufferLockingDesc);

Log(LOG_OUT,"TDirectDraw::Constructor()",">>> The Direct Draw object has been created successfully <<<");
fDebugging = false;
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::~TDirectDraw()
{
Log(LOG_IN,"TDirectDraw::Destructor()","");

// restore the original screen settings
if (fDisplayModeChanged)
	{
	Log(LOG_INFO,"TDirectDraw::Destructor()","about to restore the original display mode");
    fErrorValue = fDirectDraw->RestoreDisplayMode();
    if (fErrorValue != DD_OK)
        Error(fErrorValue,"TDirectDraw::Destructor()","display mode could not be restored");
    else
	    Log(LOG_NOTE,"TDirectDraw::Destructor()","the original display mode was restored successfully");
    }

// make sure to release all bits and pieces
ReleaseSurfaces();

// get rid of the direct draw object
if (fDirectDraw != NULL)
	{
	Log(LOG_INFO,"TDirectDraw::Destructor()","just about to release the IDirectDraw object");
    fDirectDraw->Release();
	Log(LOG_NOTE,"TDirectDraw::Destructor()","the main IDirectDraw object has been released");
    }

Log(LOG_INFO,"TDirectDraw::Destructor()","just about to delete the contents of TStringList* ErrorStrings");
Log(LOG_OUT,"TDirectDraw::Destructor()","");

if (fErrorStrings != NULL) { fErrorStrings->Clear(); delete fErrorStrings; }
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::SetEnabled(bool enabled)
{
if (enabled)
	{
    if (!fEnabled)
    	{
		Log(LOG_IN,"TDirectDraw::SetEnabled()","attempting to enable DirectDraw object");

		Log(LOG_INFO,"TDirectDraw::SetEnabled()","just about to release surfaces");
		ReleaseSurfaces();

		// make the application full screen & exclusive
        if (!fExclusiveModeSet)
        	{
			Log(LOG_INFO,"TDirectDraw::SetEnabled()","Just about to set the IDirectDraw object to exclusive / fullscreen mode");
	        fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
			if (fErrorValue != DD_OK)
				{
			    Error(fErrorValue,"TDirectDraw::SetEnabled()","Could not set the cooperative level to exclusive / fullscreen mode");
		        fEnabled = false;
			    return;
			    }
			fExclusiveModeSet = true;
			Log(LOG_NOTE,"TDirectDraw::SetEnabled()","exclusive / fullscreen mode has been set");
        	}
        else
			Log(LOG_NOTE,"TDirectDraw::SetEnabled()","exclusive / fullscreen mode has already been set");

		// set the video mode to what we're after
        if (!fDisplayModeChanged)
        	{
			Log(LOG_INFO,"TDirectDraw::SetEnabled()","just about to set the video mode to " + IntToStr(fWidth) + "x" + IntToStr(fHeight) + "x" + IntToStr(fBitsPerPixel));
			fErrorValue = fDirectDraw->SetDisplayMode( fWidth, fHeight, fBitsPerPixel );
			if (fErrorValue != DD_OK)
				{
			    Error(fErrorValue,"TDirectDraw::SetEnabled()","Could not set the video mode to " + IntToStr(fWidth) + "x" + IntToStr(fHeight) + "x" + IntToStr(fBitsPerPixel));
		        fEnabled = false;
	            return;
			    }
            fDisplayModeChanged = true;
			Log(LOG_NOTE,"TDirectDraw::SetEnabled()","the video mode has been set to " + IntToStr(fWidth) + "x" + IntToStr(fHeight) + "x" + IntToStr(fBitsPerPixel));
            }
        else
			Log(LOG_WARNING,"TDirectDraw::SetEnabled()","the video mode has already been set");

	    // create all required surfaces
		Log(LOG_INFO,"TDirectDraw::SetEnabled()","just about to recreate surfaces");
		CreateSurfaces();
		Log(LOG_NOTE,"TDirectDraw::SetEnabled()","surfaces have been recreated");

	    // we're now enabled
	 	fEnabled = true;
        }
    else
    	{
		Log(LOG_WARNING,"TDirectDraw::SetEnabled()","attempting to set enabled when already enabled");
        }
	}
else
	{
	Log(LOG_IN,"TDirectDraw::SetEnabled()","attempting to disable DirectDraw object");
	Log(LOG_INFO,"TDirectDraw::SetEnabled()","just about to release surfaces");
	ReleaseSurfaces();
	Log(LOG_NOTE,"TDirectDraw::SetEnabled()","all surfaces have been released");
    fEnabled = false;
    }
Log(LOG_OUT,"TDirectDraw::SetEnabled()","");
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetWidth(int width)
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::SetWidth()","width=" + IntToStr(width));
if (fWidth != width)
	{
	Log(LOG_INFO,"TDirectDraw::SetWidth()","Attempting to change display mode to: " + IntToStr(width) + "x" + IntToStr(fHeight) + "x" + IntToStr(fBitsPerPixel));
    if (SetDisplayMode(width,fHeight,fBitsPerPixel))
    	{
        Log(LOG_NOTE,"TDirectDraw::SetWidth()","display mode changed to: " + IntToStr(width) + "x" + IntToStr(fHeight) + "x" + IntToStr(fBitsPerPixel));
        fWidth = width;
		Log(LOG_INFO,"TDirectDraw::SetWidth()","about to recreate the required surfaces");
        }
    }
Log(LOG_OUT,"TDirectDraw::SetWidth()","");
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetHeight(int height)
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::SetHeight()","height=" + IntToStr(height));
if (fHeight != height)
	{
	Log(LOG_INFO,"TDirectDraw::SetHeight()","Attempting to change display mode to: " + IntToStr(fWidth) + "x" + IntToStr(height) + "x" + IntToStr(fBitsPerPixel));
    if (SetDisplayMode(fWidth,height,fBitsPerPixel))
    	{
        Log(LOG_NOTE,"TDirectDraw::SetHeight()","display mode changed to: " + IntToStr(fWidth) + "x" + IntToStr(height) + "x" + IntToStr(fBitsPerPixel));
        fHeight = height;
		Log(LOG_INFO,"TDirectDraw::SetHeight()","about to recreate the required surfaces");
        }
    }
Log(LOG_OUT,"TDirectDraw::SetHeight()","");
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetBitsPerPixel(int bitsperpixel)
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::SetBitsPerPixel()","bitsperpixel=" + IntToStr(bitsperpixel));
if (fBitsPerPixel != bitsperpixel)
	{
	Log(LOG_INFO,"TDirectDraw::SetBitsPerPixel()","Attempting to change display mode to: " + IntToStr(fWidth) + "x" + IntToStr(fHeight) + "x" + IntToStr(bitsperpixel));
    if (SetDisplayMode(fWidth,fHeight,bitsperpixel))
    	{
        Log(LOG_NOTE,"TDirectDraw::SetBitsPerPixel()","display mode changed to: " + IntToStr(fWidth) + "x" + IntToStr(fHeight) + "x" + IntToStr(bitsperpixel));
		fBitsPerPixel = bitsperpixel;
		Log(LOG_INFO,"TDirectDraw::SetBitsPerPixel()","about to recreate the required surfaces");
        }
    }
Log(LOG_OUT,"TDirectDraw::SetBitsPerPixel()","");
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetBackSurfaceCount(int backsurfacecount)
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::SetBackSurfaceCount()","backsurfacecount=" + IntToStr(backsurfacecount));
if (fBackBufferCount != backsurfacecount)
	{
    Log(LOG_INFO,"TDirectDraw::SetBackSurfaceCount()","about to recreate surfaces");
    ReleaseSurfaces();
    fBackBufferCount = backsurfacecount;
    CreateSurfaces();
	}
Log(LOG_OUT,"TDirectDraw::SetBackSurfaceCount()","");
}
//---------------------------------------------------------------------------
HRESULT	__fastcall	TDirectDraw::CreateSurfaces()
{
Log(LOG_IN,"TDirectDraw::CreateSurfaces()","");
if (fPrimarySurface == NULL)
	{
	// create a primary surface for this direct draw object
	Log(LOG_INFO,"TDirectDraw::CreateSurfaces()","just about to create the primary surface");
	DDSURFACEDESC	ddSurfaceDesc;
	memset(&ddSurfaceDesc,0,sizeof(ddSurfaceDesc));

	ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);
	ddSurfaceDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
	ddSurfaceDesc.dwBackBufferCount = fBackBufferCount;

	fErrorValue = fDirectDraw->CreateSurface(&ddSurfaceDesc,&fPrimarySurface,NULL);
	if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not create the primary surface");
    else Log(LOG_NOTE,"TDirectDraw::CreateSurfaces()","the primary surface has been created");

    if (fBackBuffer == NULL)
    	{
		// retrieve the back buffer
		Log(LOG_INFO,"TDirectDraw::CreateSurfaces()","just about to retrieve back buffer");
		DDSCAPS			ddsCaps;
		ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;

		fErrorValue = fPrimarySurface->GetAttachedSurface(&ddsCaps,&fBackBuffer);
		if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not retrieve back buffer");
        else Log(LOG_NOTE,"TDirectDraw::CreateSurfaces()","back buffer has been retrieved");
		}
    else
    	{
        Error(fErrorValue,"TDirectDraw::CreateSurfaces()","the back buffer exists without a primary surface");
        }
	}
else
	Log(LOG_WARNING,"TDirectDraw::CreateSurfaces()","primary surface has already been created");

Log(LOG_OUT,"TDirectDraw::CreateSurfaces()","");
return fErrorValue;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::ReleaseSurfaces()
{
Log(LOG_IN,"TDirectDraw::ReleaseSurfaces()","");
// release our back buffer(s) ???
if (fBackBuffer != NULL)
	{
	Log(LOG_INFO,"TDirectDraw::ReleaseSurfaces()","just about to release the back buffer");
	fBackBuffer->Release();
    fBackBuffer = NULL;
	Log(LOG_NOTE,"TDirectDraw::ReleaseSurfaces()","back buffer has been released");
	}
else
	Log(LOG_INFO,"TDirectDraw::ReleaseSurfaces()","Back buffer was not allocated, thus not released");

// get rid of our primary surface
if (fPrimarySurface != NULL)
	{
	Log(LOG_INFO,"TDirectDraw::ReleaseSurfaces()","just about to release the primary surface & any attached back buffers");
    fPrimarySurface->Release();
    fPrimarySurface = NULL;
	Log(LOG_NOTE,"TDirectDraw::ReleaseSurfaces()","primary surface & any attached back buffers have been released");
    }
else
	Log(LOG_INFO,"TDirectDraw::ReleaseSurfaces()","Primary surface was not allocated, thus not released");
Log(LOG_OUT,"TDirectDraw::ReleaseSurfaces()","");
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::LockSurface(LPDIRECTDRAWSURFACE surface)
{
if (!fEnabled) return false;
Log(LOG_IN,"TDirectDraw::LockSurface()","surface=" + IntToStr((int)surface));
if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Log(LOG_WARNING,"TDirectDraw::LockSurface()","surface cannot be locked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (!fBackBufferLocked)
	    	{
	        Log(LOG_INFO,"TDirectDraw::LockSurface()","about to lock the back buffer");
	        fErrorValue = surface->Lock(NULL,&fBackBufferLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the back buffer could not be locked");
		        return false;
	            }
            fBackBufferLocked = true;
			fBackBufferUsedLast = true;
            fPrimarySurfaceUsedLast = false;
		    Log(LOG_NOTE,"TDirectDraw::LockSurface()","the back buffer locked successfully");
        	}
        else
		    Log(LOG_WARNING,"TDirectDraw::LockSurface()","the back buffer is already locked");
	    }
    else
    	if (!fPrimarySurfaceLocked)
	    	{
    	    Log(LOG_INFO,"TDirectDraw::LockSurface()","about to lock the primary surface");
	        fErrorValue = surface->Lock(NULL,&fPrimarySurfaceLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the primary surface could not be locked");
		        return false;
	            }
            fPrimarySurfaceLocked = true;
            fBackBufferUsedLast = false;
			fPrimarySurfaceUsedLast = true;
		    Log(LOG_NOTE,"TDirectDraw::LockSurface()","the primary surface locked successfully");
	        }
        else
		    Log(LOG_WARNING,"TDirectDraw::LockSurface()","the primary surface is already locked");
	return true;
    }
Log(LOG_OUT,"TDirectDraw::LockSurface()","");
return false;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::UnLockSurface(LPDIRECTDRAWSURFACE surface)
{
if (!fEnabled) return false;
Log(LOG_IN,"TDirectDraw::UnLockSurface()","surface=" + IntToStr((int)surface));
if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Log(LOG_WARNING,"TDirectDraw::UnLockSurface()","surface cannot be unlocked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (fBackBufferLocked)
        	{
	        Log(LOG_INFO,"TDirectDraw::UnLockSurface()","about to lock the back buffer");
			surface->Unlock(fBackBufferLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the back buffer could not be unlocked");
		        return false;
	            }
            fBackBufferLocked = false;
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the back buffer unlocked successfully");
            }
        else
		    Log(LOG_WARNING,"TDirectDraw::UnLockSurface()","the back buffer was already unlocked");
	    }
    else
    	if (fPrimarySurfaceLocked)
	    	{
	        Log(LOG_INFO,"TDirectDraw::UnLockSurface()","about to lock the primary surface");
			fErrorValue = surface->Unlock(fPrimarySurfaceLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the primary surface could not be unlocked");
		        return false;
	            }
	        fPrimarySurfaceLocked = false;
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface unlocked successfully");
	        }
        else
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface was already unlocked");
	return true;
    }
Log(LOG_OUT,"TDirectDraw::UnLockSurface()","");
return false;
}
//---------------------------------------------------------------------------
HDC	__fastcall TDirectDraw::GetDC()
{
if (!fEnabled) return NULL;
Log(LOG_IN,"TDirectDraw::GetDC()","");

HDC hdc;

// try getting a DC for the back buffer
Log(LOG_INFO,"TDirectDraw::GetDC()","checking if we should get a DC for the back buffer");
if (!fBackBufferDCRetrieved && !fPrimarySurfaceDCRetrieved)
	{
//	Log(LOG_INFO,"TDirectDraw::GetDC()","attempting to lock the back buffer");
//    if (LockSurface(fBackBuffer))
    if (fBackBuffer != NULL)
       	{
		Log(LOG_INFO,"TDirectDraw::GetDC()","about to get a DC from the back buffer");
		fErrorValue = fBackBuffer->GetDC(&hdc);
	    if (fErrorValue != DD_OK)
	    	{
	        Error(fErrorValue,"TDirectDraw::GetDC()","Could not retrieve a DC for the back buffer");
            fHandle = NULL;
		    fBackBufferDCRetrieved = false;
	        }
        else
        	{
			fHandle = hdc;
		    fBackBufferDCRetrieved = true;
			fBackBufferUsedLast = true;
            fPrimarySurfaceUsedLast = false;
	    	Log(LOG_NOTE,"TDirectDraw::GetDC()","a DC has been retrieved for the back buffer");
			}
        }
	else
    	{	// backbuffer is null, return null
        fBackBufferDCRetrieved = false;
        Log(LOG_WARNING,"TDirectDraw::GetDC()","Could not retrieve a DC 'cause the back buffer is NULL");
        }
	}
else
	Log(LOG_WARNING,"TDirectDraw::GetDC()","either the backbuffer or the primary surface is already retrieved");

// try getting a DC for the primary surface, as this is really what GDI / DirectX wants
Log(LOG_INFO,"TDirectDraw::GetDC()","checking if we should get a DC for the primary surface");
if (!fBackBufferDCRetrieved && !fPrimarySurfaceDCRetrieved)
	{
//	Log(LOG_INFO,"TDirectDraw::GetDC()","attempting to lock the primary surface");
//    if (LockSurface(fPrimarySurface))
    if (fPrimarySurface != NULL)
       	{
		Log(LOG_INFO,"TDirectDraw::GetDC()","about to get a DC from the primary surface");
		fErrorValue = fPrimarySurface->GetDC(&hdc);
	    if (fErrorValue != DD_OK)
	    	{
	        Error(fErrorValue,"TDirectDraw::GetDC()","Could not retrieve a DC for the primary surface");
            fHandle = NULL;
	        fPrimarySurfaceDCRetrieved = false;
	        }
        else
           	{
			fHandle = hdc;
	        fPrimarySurfaceDCRetrieved = true;
            fPrimarySurfaceUsedLast = true;
			fBackBufferUsedLast = false;
			Log(LOG_NOTE,"TDirectDraw::GetDC()","a DC has been retrieved for the primary surface");
			}
		}
	else
	   	{	// primarysurface is null, return null
        fPrimarySurfaceDCRetrieved = false;
        Log(LOG_WARNING,"TDirectDraw::GetDC()","Could not retrieve a DC 'cause the primary surface is NULL");
        }
    }
else
	Log(LOG_WARNING,"TDirectDraw::GetDC()","either the backbuffer or the primary surface is already retrieved");

Log(LOG_OUT,"TDirectDraw::GetDC()","");
return fHandle;
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::ReleaseDC(HDC hdc)
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::ReleaseDC()","");
if (fBackBufferDCRetrieved)
    if (fBackBuffer != NULL)
    	{
		Log(LOG_INFO,"TDirectDraw::ReleaseDC()","about to release the back buffer DC");
		fErrorValue = fBackBuffer->ReleaseDC(&fHandle);	// ASSUMED: fBackBuffer != NULL
	    if (fErrorValue != DD_OK)
	    	{
	        Error(fErrorValue,"TDirectDraw::ReleaseDC()","Could not release a DC for the back buffer");
	        return;
	        }
	    fBackBufferDCRetrieved = false;
	    fHandle = 0;
//    Log(LOG_INFO,"TDirectDraw::ReleaseDC()","about to unlock the backbuffer");
//    UnLockSurface(fBackBuffer);
    	Log(LOG_NOTE,"TDirectDraw::ReleaseDC()","the back buffer DC has been released");
        }
	else
        Log(LOG_WARNING,"TDirectDraw::ReleaseDC()","the back buffer DC has been allocated but there is no back buffer!!");

if (fPrimarySurfaceDCRetrieved)
    if (fPrimarySurface)
		{
		Log(LOG_INFO,"TDirectDraw::ReleaseDC()","about to release the primary surface DC");
		fErrorValue = fPrimarySurface->ReleaseDC(&fHandle); // ASSUMED: fPrimarySurface != NULL
	    if (fErrorValue != DD_OK)
	    	{
	        Error(fErrorValue,"TDirectDraw::ReleaseDC()","Could not release a DC for the primary surface");
	        return;
	        }
	    fPrimarySurfaceDCRetrieved = false;
	    fHandle = 0;
//	    Log(LOG_INFO,"TDirectDraw::ReleaseDC()","about to unlock the primary surface");
//	    UnLockSurface(fPrimarySurface);
    	Log(LOG_NOTE,"TDirectDraw::ReleaseDC()","the primary surface DC has been released");
	    }
	else
        Log(LOG_WARNING,"TDirectDraw::ReleaseDC()","the primary surface DC has been allocated but there is no primary surface!!");

Log(LOG_OUT,"TDirectDraw::ReleaseDC()","");
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetDisplayMode(int width,int height,int bitsperpixel)
{
if (!fEnabled) return false;
Log(LOG_IN,"TDirectDraw::SetDisplayMode()","width=" + IntToStr(width) + ", height=" + IntToStr(height) + ", bitsperpixel=" + IntToStr(bitsperpixel) );
bool result = false;
if (fDirectDraw != NULL)
	{
	// always assume the surfaces are to be recreated
	Log(LOG_INFO,"TDirectDraw::SetDisplayMode()","just about to release surfaces");
    ReleaseSurfaces();

	Log(LOG_INFO,"TDirectDraw::SetDisplayMode()","just about to change the display mode to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
    fErrorValue = fDirectDraw->SetDisplayMode(width,height,bitsperpixel);
    if (fErrorValue != DD_OK)
    	{
        Error(fErrorValue,"TDirectDraw::SetDisplayMode()","could not set display mode to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
		result = false;
        }
    else
    	{
		Log(LOG_NOTE,"TDirectDraw::SetDisplayMode()","the display mode has been changed to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
	    fWidth = width;
	    fHeight = height;
	    fBitsPerPixel = bitsperpixel;
	    fDisplayModeChanged = true;
	    result = true;
        }

	Log(LOG_INFO,"TDirectDraw::SetDisplayMode()","just about to recreate surfaces");
    CreateSurfaces();
    }
Log(LOG_OUT,"TDirectDraw::SetDisplayMode()","");
return result;
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::RestoreDisplayMode()
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::RestoreDisplayMode()","");
if (fDirectDraw != NULL)
	{
	Log(LOG_INFO,"TDirectDraw::SetDisplayMode()","just about to restore the original display mode");
	fErrorValue = fDirectDraw->RestoreDisplayMode();
    if (fErrorValue != DD_OK)
        Error(fErrorValue,"TDirectDraw::SetDisplayMode()","could not restore the display mode");
    else
		Log(LOG_NOTE,"TDirectDraw::SetDisplayMode()","the original display mode has been restored");
    }
Log(LOG_OUT,"TDirectDraw::RestoreDisplayMode()","");
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Flip()
{
if (!fEnabled) return;
Log(LOG_IN,"TDirectDraw::Flip()","");

//
// NOTE: only surfaces with DDSCAPS_FLIP and DDSCAPS_FRONTBUFFER can be flipped
//		 ie. the primary surface
//

if (fPrimarySurface != NULL)
   	{
	Log(LOG_INFO,"TDirectDraw::Flip()","about to flip the primary surface");
    fErrorValue = fPrimarySurface->Flip(NULL,0);//DDFLIP_WAIT);
    if (fErrorValue != DD_OK)
    	{
        Error(fErrorValue,"TDirectDraw::Flip()","the primary surfacecould not be flipped");
        }
    else
		Log(LOG_NOTE,"TDirectDraw::Flip()","the primary surface has been flipped successfully");
    }
Log(LOG_OUT,"TDirectDraw::Flip()","");
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Log( int msgclass, System::AnsiString functiondesc, System::AnsiString msg )
{
if (!fDebugging) return;
switch (msgclass)
	{
    case	LOG_IN:
    		fErrorStrings->Add("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
    		if (msg != "") fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc + " --> " + msg);
            else fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc);
            break;

    case	LOG_OUT:
    		if (msg != "") fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc + " --> " + msg);
            else fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc);
    		fErrorStrings->Add("--------------------------------------------------------------------------------------");
            break;

    case	LOG_INFO:
            if (msg != "") fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc + " --> " + msg);
            else fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc);
            break;

    case	LOG_NOTE:
    		if (msg != "") fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc + " --> " + msg);
            else fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc);
            break;

    case	LOG_WARNING:
    		if (msg != "") fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "WARNING> " + functiondesc + " --> " + msg);
    		else fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "WARNING> " + functiondesc);
            break;

    case	LOG_ERROR:
    		if (msg != "") fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc + " --> " + msg);
    		else fErrorStrings->Add("#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc);
            break;
    }
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg )
{
Log(LOG_ERROR,functiondesc,"Err # " + IntToStr(errorvalue) + " " + ErrorMessage(errorvalue));
Log(LOG_ERROR,functiondesc,"Err : " + errormsg);
}
//---------------------------------------------------------------------------
System::AnsiString	__fastcall	TDirectDraw::ErrorMessage(int errorvalue)
{
System::AnsiString	errormessage = "";
switch (errorvalue)
	{
    // common directx results
    case DD_OK:						errormessage = "DD_OK, DC retrieved successfully"; break;
	case DDERR_GENERIC:             errormessage = "DDERR_GENERIC, generic error"; break;
	case DDERR_INVALIDOBJECT:		errormessage = "DDERR_INVALIDOBJECT, Invalid object"; break;
	case DDERR_INVALIDPARAMS:		errormessage = "DDERR_INVALIDPARAMS, Invalid params"; break;
	case DDERR_INVALIDSURFACETYPE:	errormessage = "DDERR_INVALIDSURFACETYPE, Invalid surface type"; break;
	case DDERR_UNSUPPORTED:			errormessage = "DDERR_UNSUPPORTED, Unsupported"; break;
	case DDERR_SURFACELOST:			errormessage = "DDERR_SURFACELOST, Surface is lost"; break;
	case DDERR_SURFACEBUSY:			errormessage = "DDERR_SURFACEBUSY, Surface is busy"; break;
	case DDERR_WASSTILLDRAWING:		errormessage = "DDERR_WASSTILLDRAWING, Was still drawing"; break;
	// flip results
	case DDERR_NOFLIPHW:			errormessage = "DDERR_NOFLIPHW, No flip hardware"; break;
	case DDERR_NOTFLIPPABLE:		errormessage = "DDERR_NOTFLIPPABLE, Surface is not flippable"; break;
    // GetDC results
	case DDERR_DCALREADYCREATED:	errormessage = "DDERR_DCALREADYCREATED, DC was already created"; break;
    default:						errormessage = "UNKNOWN ERROR, the error generated is unknown";
    };
return errormessage;
}
//---------------------------------------------------------------------------
/*void	__fastcall	TDirectDraw::Clear(TColor color)
{
Log(LOG_IN,"TDirectDraw::Clear()","color="+IntToStr(color));
if (fBackBufferUsedLast)
	{
	Log(LOG_INFO,"TDirectDraw::Clear()","about to clear the back buffers memory");
	memset( fBackBufferLockingDesc.lpSurface, color, fWidth * fHeight );
	Log(LOG_INFO,"TDirectDraw::Clear()","about to flip the back buffer");
	Flip();
    }
else
    if (fDrawingOnPrimarySurface)
		{
		Log(LOG_INFO,"TDirectDraw::Clear()","about to clear the primary surfaces memory");
		memset( fPrimarySurfaceLockingDesc.lpSurface, color, fWidth * fHeight );
		Log(LOG_INFO,"TDirectDraw::Clear()","about to flip the primary surface");
        Flip();
	    }
Log(LOG_OUT,"TDirectDraw::Clear()","");
}
//---------------------------------------------------------------------------
// tell fDirectDraw to set fullscreen / exclusive mode
fErrorValue = DirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
if (fErrorValue != DD_OK)
	{
    fErrorStrings->Add("Error: Could not set the cooperative level");
    return false;
    }

// set the display mode to 640x480x256
fErrorValue = DirectDraw->SetDisplayMode( 640, 480, 8 );
if (fErrorValue != DD_OK)
	{
    fErrorStrings->Add("Error: Could not set the display mode");
    return false;
    }

// all done
return true
*/
