//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TDirectDrawForm.h"
//---------------------------------------------------------------------------
#pragma link "TDirectDraw"
#pragma link "Grids"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TDirectDrawForm *DirectDrawForm;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// remember: 	SetPriorityClass(hProcess: THandle; dwPriorityClass: DWORD): BOOL; stdcall;
// remember:	Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TDirectDrawForm::TDirectDrawForm(TComponent* Owner)
	: TForm(Owner)
{
fAnimationTicks = 0;
fFramesFlipped = 0;
fSeconds = 0;
fColor = 0;

fDirectXDC = NULL;
fHandlingWMPAINT = false;
fControlBeingPainted = -1;

fDirectDraw = new TDirectDraw(this);
if (fDirectDraw == NULL)
	{
	fDirectDrawCreated = false;
    }
else
	{
	fDirectDrawCreated = true;
    DirectDraw->Debugging = false;
	}
}
//---------------------------------------------------------------------------
__fastcall 	TDirectDrawForm::~TDirectDrawForm()
{
if (fDirectDraw != NULL) delete fDirectDraw;
}
/*//---------------------------------------------------------------------------
void 	__fastcall TDirectDrawForm::CMInvalidate(Messages::TMessage &Message)
{
if (fDirectDrawCreated)
	{
    if (!DirectDraw->Enabled)
    	{
        TForm::CMInvalidate(Message);
        return;
        }
    }
else
	{
    TForm::CMInvalidate(Message);
    return;
    }

DirectDraw->Log(LOG_NOTE,"<<FORM>>::CMInvalidate()","recieved CM_INVALIDATE");
//    if (Parent <> NULL) Parent->Perform(CM_INVALIDATE, 1, false);
if (Message.WParam == 0) InvalidateRect(Handle, NULL, false);
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::InvalidateFrame()
{
if (fDirectDrawCreated)
	{
    if (!DirectDraw->Enabled)
    	{
        TForm::InvalidateFrame();
        return;
        }
    }
else
	{
    TForm::InvalidateFrame();
    return;
    }

RECT r;
r.left = BoundsRect.Left;
r.top = BoundsRect.Top;
r.right = BoundsRect.Right;
r.bottom = BoundsRect.Bottom;
InflateRect(&r, 1, 1);
InvalidateRect(Handle, &r, false);
} */
//---------------------------------------------------------------------------
void 	__fastcall TDirectDrawForm::WndProc(Messages::TMessage &Message)
{
if (fDirectDrawCreated)
	{
    if (!DirectDraw->Enabled)
    	{
        TForm::WndProc(Message);
        return;
        }
    }
else
	{
    TForm::WndProc(Message);
    return;
    }

// handle any repaint messages coming from Windows rather than letting the default
// WndProc function handle them as the default WndProc function automatically draws
// the form when receiving WM_PAINT, WM_NCPAINT, WM_NCACTIVATE.....
// since the form redrawing code uses a windows gdi Paint3d function which blits to
// the display on completion (we have no control over this), it is probably better
// to use our own handler so that we can control exactly when to blit what.

LPDRAWITEMSTRUCT lpdraw;
int		message = Message.Msg;

switch (message)
	{
    case WM_PAINT:
    case WM_NCPAINT:

		// if we're possibly using DirectDraw stuff & we're not already handling a paint message
    	if (fDirectDrawCreated && !fHandlingWMPAINT)
        	// and DirectDraw stuff is enabled
            if (DirectDraw->Enabled)
	            {
		    	// simply log all WM_PAINT & NC_PAINT messages for posterity
		    	if (message == WM_PAINT) DirectDraw->Log(LOG_IN,"<<FORM>>::WndProc()","Received WM_PAINT, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
		        else DirectDraw->Log(LOG_IN,"<<FORM>>::WndProc()","Received WM_NCPAINT, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));

                // set a flag so that we don't handle multiple WM_PAINT messages while
                // we're still handling the first WM_PAINT message (this can happen)
	            fHandlingWMPAINT = true;

		        // get a relevant DC from DirectDraw, currently TDirectDraw returns
                // the backbuffer if possible, primary surface otherwise
	            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->GetDC()");
		      	fDirectXDC = DirectDraw->GetDC(false);
//                fDirectXDC = DirectDraw->GetBackBufferDC();
	            DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->GetDC()");

    			// if the DC retrieved is a valid DC
		        if (fDirectXDC != NULL)
		           	{
		            // tell windows that the entire region of the form will need to be repainted
		            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","InvalidateRect()");
		            InvalidateRect(Handle,NULL,false);
	    	        DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","InvalidateRect()");

	                // call our overloaded PaintHandler() with appropriate message
	                DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","PaintHandler()");
	                Messages::TWMPaint	MyPaintMessage;
	                MyPaintMessage.Msg = WM_PAINT;
	                MyPaintMessage.DC  = fDirectXDC;
	                MyPaintMessage.Unused = 0;
	                MyPaintMessage.Result = 0;
	                PaintHandler(MyPaintMessage);
	                DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","PaintHandler()");

				    // release the DirectX supplied DC
					DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->ReleaseDC()");
					DirectDraw->ReleaseDC(fDirectXDC);
	                fDirectXDC = NULL;
					DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->ReleaseDC()");

		            // tell windows that the entire region of the form has been repainted
		            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","ValidateRect()");
		            ValidateRect(Handle,NULL);
	    	        DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","ValidateRect()");

					// flip the DirectX buffers (makes recent drawing visible to the user).
					DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->FlipBackBufferToGDI()");
	                DirectDraw->FlipBackBufferToGDI();
					DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->FlipBackBufferToGDI()");

	                // note that directx drawing has worked successfully
					DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","!! Direct Drawing Worked !!");
		            }

                // reset the flag used to stop multiple WM_PAINT messages while
                // we're still handling the first WM_PAINT message (this can happen)
		        fHandlingWMPAINT = false;
	            DirectDraw->Log(LOG_OUT,"<<FORM>>::WndProc()","");

		        // clear the Message so that the VCL does nothing else & call the VCL stuff
		        Message.Msg = 0;
                return;
	            }
		// drop through and let the VCL handle the message
        break;

	case WM_DRAWITEM:

/*		// only modify behaviour when using DirectDraw stuff, otherwise drop through to
        // the VCL default WndProc()
    	if (fDirectDrawCreated)
        	{
            DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_DRAWITEM, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
            // if we're in the middle of a WM_PAINT message, let through the VCL do the work
	    	if (fHandlingWMPAINT)
            	{
	    		DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","WM_PAINT in progress, WM_DRAWITEM passed through to VCL");
			    TForm::WndProc(Message);
                return;
                }
            // otherwise if DirectDraw stuff is enabled, simply clear WM_DRAWITEM messages
	        else
            	if (DirectDraw->Enabled)
		        	{
		    		DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","WM_DRAWITEM received while not handling WM_PAINT, ignored");
		        	Message.Msg = 0;
                    return;
		            }
            }
		// drop through and let the VCL handle the message
*/
//		lpdraw = (LPDRAWITEMSTRUCT) Message.LParam;
//		if (lpdraw->CtlType == ODT_LISTBOX) return;
		break;

   	// don't log any messages for WM_MEASUREITEM messages
    case WM_MEASUREITEM:

		break;

    case WM_VSCROLL:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_VSCROLL, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
/*    	if (fDirectDrawCreated)
        	{
            if (fHandlingWMPAINT)
            	{
                // don't do anything special
//                }
  //          else
    //            if (DirectDraw->Enabled)
	  //          	{
                    // force the scrollbars to repaint
	                HWND hwnd = (HWND) Message.LParam;
                    int vpos = GetScrollPos(hwnd, SB_VERT);
                    int hpos = GetScrollPos(hwnd, SB_HORZ);
                    int cpos = GetScrollPos(hwnd, SB_CTL);
                    int pos = 0, flags = 0;

                    if (vpos != 0) { pos = vpos; flags = SB_VERT; }
                    else if (hpos != 0) { pos = hpos; flags = SB_HORZ; }
                    else if (cpos != 0) { pos = cpos; flags = SB_CTL; }

				    if (flags != 0)
                    	{
                        SetScrollPos(hwnd, flags, pos+1, false);
	                    SetScrollPos(hwnd, flags, pos, true);
                        }
                    }
            Message.Msg = 0;
            return;
            }
*/        break;

    case WM_HSCROLL:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_HSCROLL, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
    	break;

    // simply log the various messages for debugging purposes
	case WM_COMMAND:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_COMMAND, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

    // simply log the various messages for debugging purposes... etc.
	case WM_GETTEXT:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_GETTEXT, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_GETTEXTLENGTH:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_GETTEXTLENGTH, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_MOUSEFIRST:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_MOUSEFIRST, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case CN_DRAWITEM:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received CN_DRAWITEM, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
		break;

	case WM_CTLCOLORMSGBOX:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLORMSGBOX, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_CTLCOLOREDIT:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLOREDIT, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_CTLCOLORLISTBOX:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLORLISTBOX, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_CTLCOLORBTN:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLORBTN, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_CTLCOLORDLG:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLORDLG, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_CTLCOLORSCROLLBAR:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLORSCROLLBAR, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_CTLCOLORSTATIC:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_CTLCOLORSTATIC, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

	case WM_SETCURSOR:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_SETCURSOR, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

    case WM_NCACTIVATE:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_NCACTIVATE, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
    	break;

	case WM_NCHITTEST:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","received WM_NCHITTEST, WParam=" + IntToStr(Message.WParam) + ", LParam="+IntToStr(Message.LParam));
        break;

    default:

    	if (fDirectDrawCreated) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","unknown message " + IntToStr(message));
        break;
    }

// let the VCL handle the message
TForm::WndProc(Message);
Message.Msg = 0;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWindow(HDC hdc)
{
if (hdc == NULL) return;
if (fDirectDrawCreated)
	{
	if (!DirectDraw->Enabled)
    	{
	    TForm::PaintWindow(hdc);
	    return;
	    }
    }
else
	{
	TForm::PaintWindow(hdc);
    return;
    }

DirectDraw->Log(LOG_IN,"<<FORM>>::PaintWindow()","");

try
	{
    Canvas->Handle = hdc;
    Paint();
    }
catch ( ... )
	{
    }
// always release anything that might have been allocated in Paint()
Canvas->Handle = NULL;

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintWindow()","");
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintControls(HDC hdc, TControl* First)
{
// simply check our preconditions & skip out if not appropriate
if (hdc == NULL) return;
if (fDirectDrawCreated)
	{
	if (!DirectDraw->Enabled)
    	{
	    TForm::PaintControls(hdc,First);
	    return;
	    }
    }
else
	{
	TForm::PaintControls(hdc,First);
    return;
    }
    
DirectDraw->Log(LOG_IN,"<<FORM>>::PaintControls()","");
if (ControlCount == 0)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","ControlCount is 0");
    return;
    }

// quickly scan the controls until we get to the first control, then fall through with the
// value of the first control in "i" (defaulting at zero if First parameter is NULL)
// ; also determine how many controls are meant to be repainted
int controls_to_paint = ControlCount;
int i = 0;
int j = 0;
if (First != NULL)
	{
    controls_to_paint = 1;
	while (j<ControlCount)
	   	{
		if (Controls[j] == First) i = j;
	    j++;
	    }
    }

// create a rectangular region used for resetting the windows clip region (after each
// control is repainted, the clip region is set to that control's clip region - thus we
// reset it to the form's clip region)

HRGN whole_form_clip_rect = CreateRectRgn(Left,Top,Width,Height);
GetClipRgn(hdc,&whole_form_clip_rect);

// set the clip region to that of the form....
SelectClipRgn(hdc,whole_form_clip_rect);
// set the window origin to 0,0 (caJUSTse)
POINT p;
p.x = 0;
p.y = 0;
SetWindowOrgEx( hdc, 0, 0, &p );


// scan through all controls, starting with the first of the ordered controls and get them
// to paint to the DirectX supplied DC

int k = 0;
while (k<controls_to_paint)
	{
    // set a global flag used by the WndProc() function
	fControlBeingPainted = i;

    // use mapped variables
    TControl* control = Controls[i];

    // figure out if the class is based on TWinControl (if so, it may require a border).
    // this is done by traversing the ancestor classes of the control until we find a
    // TWinControl parent, or till we run out of parents
    TClass	theclass = control->ClassType();
    bool 	isWinControl = false;
    bool	isScrollingWinControl = false;
    bool	isListBox = false;
	while	(theclass != NULL)
		{
	    System::AnsiString	class_name = theclass->ClassName();
//		DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' ancestors - '" + class_name + "'");
        if (class_name == "TListBox")
        	{
            isListBox = true;
            break;
            }
        if (class_name == "TScrollingWinControl")
        	{
            isScrollingWinControl = true;
            break;
            }
        if (class_name == "TWinControl")
        	{
            isWinControl = true;
            break;
            }
		theclass = theclass->ClassParent();
		}

    // if we _really_ need to redraw the control
   	if (control->Visible && !control->ControlStyle.Contains(csNoDesignVisible))
    	{
        // debug
		DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Painting Control - '" + control->Name + "' of class - '" + control->ClassName() + "'");

		// figure out where we should be drawing the relevant control
		int borderx = GetSystemMetrics(SM_CXDLGFRAME);
		int bordery = GetSystemMetrics(SM_CYDLGFRAME);
        int captionheight = GetSystemMetrics(SM_CYCAPTION);

        // when required, draw a simple frame around the control....
        // note: clip rect is entire form at this point
		bool control_is_transparent = !control->ControlStyle.Contains(csOpaque);
        bool control_is_framed = control->ControlStyle.Contains(csFramed);

        // set the clip rect to that of the control
		RECT r2;
		r2.left = Left + borderx + control->Left;
		r2.top  = Top + bordery + captionheight + control->Top;
		r2.right  = r2.left + control->Width + 1;
		r2.bottom = r2.top + control->Height + 1;
		HRGN inside_border_clip_rect = CreateRectRgn( r2.left, r2.top, r2.right, r2.bottom );
		SelectClipRgn( hdc, inside_border_clip_rect );
        DeleteObject( inside_border_clip_rect );

        // set the window origin to the top,left of the control to be drawn....
        int movewindow_dx = r2.left;
        int movewindow_dy = r2.top;
        MoveWindowOrg( hdc, movewindow_dx, movewindow_dy);

// draw debugging lines
//RECT nr1;
//nr1.left = 0;
//nr1.top = 0;
//nr1.right = control->Width;
//nr1.bottom = control->Height;
//HBRUSH noticable2 = CreateSolidBrush(ColorToRGB(clYellow));
//FrameRect(hdc, &nr1, noticable2);
//DeleteObject( noticable2 );

		// paint the control using various control methods, depending on the type of control
        if (isListBox)
        	{
            // simply tell the control where to paint
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()",control->Name + " is a TListBox");
		    TListBox* ListBox = (TListBox*) control;
            ListBox->PaintTo(hdc,0,0);

	        // log a quick note
			DirectDraw->Log(LOG_NOTE,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - was painted using ->PaintTo()");
            }
        else
        if (isScrollingWinControl)
        	{
            // simply tell the control where to paint
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()",control->Name + " is a TScrollingWinControl");
		    TScrollingWinControl* ScrollingWinControl = (TScrollingWinControl*) control;
            ScrollingWinControl->PaintTo(hdc,0,0);

	        // log a quick note
			DirectDraw->Log(LOG_NOTE,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - was painted using ->PaintTo()");
            }
        else if (isWinControl)
        	{
            // simply tell the control where to paint
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()",control->Name + " is a TWinControl");
		   	TWinControl* WinControl = (TWinControl*) control;
            WinControl->PaintTo(hdc,0,0);
	        // log a quick note
			DirectDraw->Log(LOG_NOTE,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - was painted using ->PaintTo()");
            }
        else
        	{
		    // only clear the background for controls that are transparent... little bit of
	        // a speed increase hopefully
            if (!control->ControlStyle.Contains(csOpaque))
            	{
				DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()",control->Name + "->Perform(WM_ERASEBKGND)");
                control->Perform(WM_ERASEBKGND, (long) hdc, NULL);
                }

	        // tell the control to repaint itself onto the DirectX supplied DC
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()",control->Name + "->Perform(WM_PAINT)");
	      	control->Perform(WM_PAINT, (long) hdc, NULL);

	        // log a quick note
			DirectDraw->Log(LOG_NOTE,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - was painted using ->Perform(WM_PAINT)");
            }

        //for (int adelay=0;adelay<30000000;adelay++) ;

        // restore the original window origin
        MoveWindowOrg( hdc, -movewindow_dx, -movewindow_dy );
        
  		// reset the clip region to what is was before painting this control
		SelectClipRgn( hdc, whole_form_clip_rect );

        if (isWinControl && control_is_framed)
        	{
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - has a border, painting border now");

            // paint the appropriate frame
            RECT r1;
	        HBRUSH shadow = CreateSolidBrush(ColorToRGB(clBtnShadow));
	        HBRUSH highlight = CreateSolidBrush(ColorToRGB(clBtnHighlight));

            r1.left = r2.left-1;//control->Left-1;
            r1.top = r2.top-1;//control->Top-1;
            r1.right = r2.right;//control->Left + control->Width;
            r1.bottom = r2.bottom;//control->Top + control->Height;
	        FrameRect(hdc, &r1, shadow);

            r1.left++;
            r1.top++;
            r1.right++;
            r1.bottom++;
	        FrameRect(hdc, &r1, highlight);

			DeleteObject(highlight);
            DeleteObject(shadow);
	        }
        }

	// choose the next control
    ++i %= ControlCount;	// wrap to 0..ControlCount-1
    k++;
    }

// reset the global flag used in the WndProc() function.
fControlBeingPainted = -1;

// make sure to blow away our whole_form_clip_rect
DeleteObject( whole_form_clip_rect );

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","painted " + IntToStr(k) + " controls");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintHandler(Messages::TWMPaint &Message)
{
// simply check our preconditions & skip out if not appropriate
if (fDirectDrawCreated)
	{
	if (!DirectDraw->Enabled)
    	{
	    TForm::PaintHandler(Message);
	    return;
	    }
    }
else
	{
	TForm::PaintHandler(Message);
    return;
    }

bool drawing_worked = false;
DirectDraw->Log(LOG_IN,"<<FORM>>::PaintHandler()","");

// if the supplied DC is okay
HDC hdc = Message.DC;
// use standard function calls to paint to our DirectX surface
try
  	{
    // paint the form (custom)
    DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintForm()");
    PaintForm(hdc);
//int adelay;
//for (adelay=0;adelay<30000000;adelay++) ;
    DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintForm()");

	DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintWindow()");
	PaintWindow(hdc);
//for (adelay=0;adelay<30000000;adelay++) ;
	DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintWindow()");

	DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintControls()");
	PaintControls(hdc,NULL);
//for (adelay=0;adelay<30000000;adelay++) ;
	DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintControls()");

    drawing_worked = true;
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - SUCCESSFUL!");
	}
// unless a problem occurs
catch (...)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - NOT SUCCESSFUL!");
	}
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::Repaint()
{
if (fDirectDrawCreated)
	{
	if (!DirectDraw->Enabled)
    	{
        TForm::Repaint();
        return;
        }
    }
else
	{
	TForm::Repaint();
    return;
    }
    
if (Visible)
	{
    InvalidateRect(Handle,NULL,false);
    UpdateWindow(Handle);
    }
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::PaintForm(HDC hdc)
{
// skip out if invalid DC
if (hdc == NULL) return;
if (!fDirectDrawCreated) return;

DirectDraw->Log(LOG_IN,"<<FORM>>::PaintForm()","");

// set the canvas to paint on the DC
Canvas->Handle = hdc;

// - debugging - fill the entire form area with black
//TRect	dr = Canvas->ClipRect;
//Canvas->Brush->Color = clYellow;
//Canvas->Brush->Style = bsSolid;
//Canvas->FillRect( dr );

// get some stats from windows
int borderx = GetSystemMetrics(SM_CXDLGFRAME);
int bordery = GetSystemMetrics(SM_CYDLGFRAME);
int captionheight = GetSystemMetrics(SM_CYCAPTION);
int x,y;

// draw a 3d frame round the entire form
Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;

// left & right edges
for (x=0;x<borderx;x++)
	{
	if (x==0) Canvas->Pen->Color = clBtnHighlight;
    else Canvas->Pen->Color = clBtnFace;
	Canvas->MoveTo(x,0+x);
	Canvas->LineTo(x,Height+bordery+1);
	Canvas->MoveTo(Width-x-1,0+x);
	Canvas->LineTo(Width-x-1,Height+bordery+1);
    }
// top & bottom edges
for (y=0;y<bordery;y++)
    {
	if (y==0) Canvas->Pen->Color = clBtnHighlight;
    else Canvas->Pen->Color = clBtnFace;
    Canvas->MoveTo(0+y,y);
    Canvas->LineTo(Width-y,y);
    Canvas->MoveTo(0+y,Height-y+bordery+1);
    Canvas->LineTo(Width-y,Height-y+bordery+1);
    }

// fill in the caption area with the active caption color
TRect	r = Canvas->ClipRect;
TRect	caption_r = r;
caption_r.Left += borderx;
caption_r.Top  += bordery;
caption_r.Right -= borderx;
caption_r.Bottom = caption_r.Top + captionheight;
Canvas->Brush->Color = clActiveCaption;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( caption_r );

// draw line between caption bar and main draw area
Canvas->Pen->Color = clBtnFace;
Canvas->Pen->Width = bordery;
Canvas->MoveTo(caption_r.Left, caption_r.Bottom + 1);
Canvas->LineTo(caption_r.Right, caption_r.Bottom + 1);

// draw any buttons on the caption bar



// draw the form's caption
Canvas->Brush->Color = clActiveCaption;
Canvas->Font->Size = 12;
Canvas->Font->Color = Font->Color;
Canvas->Font->Name = Font->Name;
Canvas->TextOut( caption_r.Left+borderx,caption_r.Top, Caption );
if (fDirectDrawCreated)
	{
    if (DirectDraw->Enabled)
    	{
		Canvas->Font->Color = clGreen;
		Canvas->TextOut( caption_r.Left+600,caption_r.Top, "DirectX is Enabled" );
        }
    else
    	{
		Canvas->Font->Color = clRed;
		Canvas->TextOut( caption_r.Left+600,caption_r.Top, "DirectX is Not Enabled" );
        }
    }
else
	{
	Canvas->Font->Color = clRed;
	Canvas->TextOut( caption_r.Left+600,caption_r.Top, "DirectX is Not Available" );
    }

// fill the main form area with it's color
TRect	main_r;
main_r.Left = caption_r.Left;
main_r.Top  = caption_r.Bottom + bordery;
main_r.Right = caption_r.Right;
main_r.Bottom = r.Bottom - bordery;
Canvas->Brush->Color = Color;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( main_r );

// always release anything that may have been allocated by the Canvas
Canvas->Handle = NULL;

// make sure that the window origin is inside the borders / caption area
MoveWindowOrg(hdc, main_r.Left, main_r.Top);

// make sure that the clip rectangle is inside the borders / caption area (0,0 since origin has moved)
HRGN inside_border_clip_rect = CreateRectRgn(0,0,main_r.Right-main_r.Left,main_r.Bottom-main_r.Top);
SelectClipRgn( Canvas->Handle, inside_border_clip_rect );
DeleteObject( inside_border_clip_rect );

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintForm()","");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ExitButtonClick(TObject *Sender)
{
if (Enabled) Close();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ResetCountersClick(TObject *Sender)
{
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::EnableDirectXButtonClick(TObject *Sender)
{
if (fDirectDrawCreated)
	{
    DirectDraw->Enabled = !DirectDraw->Enabled;
//    if (DirectDraw->Enabled)
//    	ShowCursor(false);
//    else
//    	ShowCursor(true);
    }
Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::UpdateDirectXMessages()
{
if (fDirectDrawCreated)
	{
    if (DirectDraw->Enabled)
		if (DirectDraw->ErrorStrings)
			{
			DebugMessages->Items->Clear();
			for (int i=0;i<DirectDraw->ErrorStrings->Count;i++)
				DebugMessages->Items->Add(DirectDraw->ErrorStrings->Strings[i]);
            DebugMessages->Repaint();
		    }
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormActivate(TObject *Sender)
{
bool all_enabled = true;
DirectDraw->Enabled = true;
if (fDirectDrawCreated)
	{
	if (DirectDraw->Enabled)
		{
		if (DirectDraw->SetDisplayMode(640,480,8))
	    	{
			DirectDraw->Log(LOG_NOTE,"",">>>>>>>>>>>>>> Display Mode Set to 640x480x8");
		    AnimationTimer->Enabled = true;
	        }
        }
    }

if (!AnimationTimer->Enabled)
	{
    for (int i=0;i<ControlCount;i++) Controls[i]->Enabled = false;
    AnimationTimer->Enabled = false;
    }

Left=0; Width=DirectDraw->Width-Left;
Top=5; Height=DirectDraw->Height-Top;
DebugMessages->Height -= 5;

fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;

// turn off the normal mouse cursor
//ShowCursor( false );
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormDeactivate(TObject *Sender)
{
AnimationTimer->Enabled = false;
DirectDraw->Enabled = false;
DebugMessages->Items->Clear();

// turn on the normal mouse cursor
//ShowCursor( true );
}
//---------------------------------------------------------------------------
// yuk, temporary debugging vars, etc.
//---------------------------------------------------------------------------
TDateTime last_time = Time();
int PageNumber;
int x = 0;
int y = 0;
int xincr = 1;
int yincr = 1;
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormPaint(TObject *Sender)
{
// --------- debugging display ----------
Canvas->Pen->Color = clRed;
Canvas->Font->Color = clWhite;
Canvas->TextOut(200,0,"PAGE # " + IntToStr(PageNumber));
if (fDirectDrawCreated) if (DirectDraw->UsingPrimarySurface) Canvas->TextOut(400,0,"Viewing: Primary Surface");
if (fDirectDrawCreated) if (DirectDraw->UsingBackBuffer) Canvas->TextOut(400,0,"Viewing: Back Buffer");

// --------- normal FormPaint() drawing code using the VCL or GDI with Canvas->Handle as the DC

// display the number of frames, etc. being drawn
TDateTime current_time = Time();
unsigned short chour,cmin,csec,cmsec;
unsigned short lhour,lmin,lsec,lmsec;
current_time.DecodeTime(&chour,&cmin,&csec,&cmsec);
last_time.DecodeTime(&lhour,&lmin,&lsec,&lmsec);
int seconds_elapsed = 0;
if (csec != lsec)
	{
    if (csec < lsec) seconds_elapsed += 60;
    seconds_elapsed += (csec - lsec);
    }
fSeconds += seconds_elapsed;
last_time = current_time;

++PageNumber %= 2;	// 0..1 only

TicksLabel->Caption = "Ticks: " + IntToStr(fAnimationTicks);
FramesLabel->Caption = "Frames: " + IntToStr(fFramesFlipped);
SecondsLabel->Caption = "Seconds: " + IntToStr(fSeconds);
if (fSeconds != 0)
	{
    FramesPerSecondLabel->Caption = "FPS: " + FloatToStrF((float)fFramesFlipped / (float)fSeconds, ffFixed, 6, 2);
	}
AnimSpeedLabel->Caption = "Anim: " + IntToStr(AnimationTimer->Interval);

for (int z=0; z<1000; z++)
   	Canvas->Pixels[random(Width)][random(Height)] = z << 16 + z << 8 + z;

// also display some DirectX stats, etc.
if (fDirectDrawCreated)
	{
    DirectXInstalledLabel->Font->Color = clGreen;
    DirectXInstalledLabel->Caption = "DirectX is Installed";

	if (DirectDraw->Enabled)
		{
	    DirectXEnabledLabel->Font->Color = clGreen;
	    DirectXEnabledLabel->Caption = "DirectX drawing is enabled";
	    }
	else
		{
	    DirectXEnabledLabel->Font->Color = clRed;
	    DirectXEnabledLabel->Caption = "DirectX drawing is not enabled.";
	    }
	if (DirectDraw->Debugging)
		{
	    DirectXDebuggingLabel->Font->Color = clGreen;
	    DirectXDebuggingLabel->Caption = "DirectX debugging is enabled";
	    }
	else
		{
	    DirectXDebuggingLabel->Font->Color = clRed;
	    DirectXDebuggingLabel->Caption = "DirectX debugging is not enabled";
	    }
    }
else
	{
    DirectXInstalledLabel->Font->Color = clRed;
    DirectXInstalledLabel->Caption = "DirectX is Not Installed";
    }

// display a box moving about the screen (note: x,y,xincr,yincr global & just above function)
Canvas->Pen->Color = (TColor) RGB(255,255,255);
Canvas->Brush->Color = (TColor) RGB((x*128)%256,(x*16)%256,(x*16+16*fColor)%256) ;
Canvas->Brush->Style = bsSolid;
Canvas->Rectangle( x,y, x+400,y+200 );

int xx = (x*2>Width/2) ? x*2 - Width/2 : Width/2 - x*2;
int yy = (y>Width/2) ? y - Height/2: Height/2 - y;
Image1->Visible = false;
Image1->SetBounds(xx,yy,x/64+Image1->Width,y/64+Image1->Height);
Image1->Visible = true;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::RepaintMessagesButtonClick(TObject *Sender)
{
// make sure the scroll box of directx messages is realtime.....
UpdateDirectXMessages();
if (DebugMessages->Showing && (ActiveControl == DebugMessages))
	{
    DebugMessages->Hide();
    DebugMessages->Show();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ClearMessagesButtonClick(TObject *Sender)
{
if (Enabled && fDirectDrawCreated)
	{
    DirectDraw->ErrorStrings->Clear();
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ToggleDebuggingButtonClick(TObject *Sender)
{
if (Enabled && fDirectDrawCreated)
	{
    DirectDraw->Debugging = !DirectDraw->Debugging;
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::AnimationTimerTimer(TObject *Sender)
{
fAnimationTicks++;
fColor++;
x += xincr;
y += yincr;

if (x<0) xincr = 1;
if (y<0) yincr = 1;
if (x+400>Width) xincr = -1;
if (y+200>Height) yincr = -1;

fFramesFlipped++;
Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn2Click(TObject *Sender)
{
if (AnimationTimer->Interval>1000) AnimationTimer->Interval -= 1000;
else if (AnimationTimer->Interval>100) AnimationTimer->Interval -= 100;
else if (AnimationTimer->Interval>10) AnimationTimer->Interval -= 10;
else if (AnimationTimer->Interval>1) AnimationTimer->Interval -= 1;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn4Click(TObject *Sender)
{
if (AnimationTimer->Interval<10) AnimationTimer->Interval += 1;
else if (AnimationTimer->Interval<100) AnimationTimer->Interval += 10;
else if (AnimationTimer->Interval<1000) AnimationTimer->Interval += 100;
else if (AnimationTimer->Interval<10000) AnimationTimer->Interval += 1000;
}
//---------------------------------------------------------------------------
