//---------------------------------------------------------------------------
#ifndef TDirectDrawH
#define TDirectDrawH
//---------------------------------------------------------------------------
#include <vcl\SysUtils.hpp>
#include <vcl\Controls.hpp>
#include <vcl\Classes.hpp>
#include <vcl\Forms.hpp>
//---------------------------------------------------------------------------
#include <d3drmwin.h>
#include <direct.h>
//---------------------------------------------------------------------------
#define		Delay(N)		for (int adelay(N)=0;adelay(N)<(N)*1000000;adelay(N)++) ; 
//---------------------------------------------------------------------------
#define	LOG_IN			0
#define	LOG_OUT			1
#define	LOG_INFO		2
#define	LOG_NOTE		3
#define	LOG_WARNING		4
#define	LOG_ERROR		5
#define	LOG_MEMSTATUS	6
#define	LOG_BEFORE		7
#define	LOG_AFTER		8
//---------------------------------------------------------------------------
class TDirectDraw : public TComponent
{
private:
    HRESULT							fErrorValue;
    int								fErrorCount;
    TStringList*					fErrorStrings;

	LPDIRECTDRAW					fDirectDraw;
    LPDIRECTDRAWSURFACE				fPrimarySurface;
    LPDIRECTDRAWSURFACE				fBackBuffer;

    HDC								fBackBufferDC;
    HDC								fPrimarySurfaceDC;
    int								fWidth;
    int								fHeight;
    int								fBitsPerPixel;
    int								fBackBufferCount;

    bool							fExclusiveModeSet;
    bool							fDisplayModeChanged;

    bool							fDebugging;
    bool							fEnabled;

    POINT							fMouseCursorPos;

    // functions used internally
    bool				__fastcall	CreateDirectDraw();
    bool				__fastcall	ReleaseDirectDraw();
    bool				__fastcall	SetExclusiveMode();
    bool				__fastcall	SetNormalMode();
    void				__fastcall	SetEnabled(bool enabled);			
    bool				__fastcall	RestoreDisplayMode();
	bool				__fastcall	CreateSurfaces();
    bool				__fastcall	ReleaseSurfaces();

    // property functions
	void				__fastcall	SetWidth(int width);
    void				__fastcall	SetHeight(int height);
	void				__fastcall	SetBitsPerPixel(int bitsperpixel);
	void				__fastcall	SetBackSurfaceCount(int backsurfacecount);

protected:

	void				__fastcall	InMemoryStatus(MEMORYSTATUS* in_mstat);
    void				__fastcall	OutMemoryStatus(MEMORYSTATUS* out_mstat);
    bool				__fastcall	SameMemoryStatus(MEMORYSTATUS* in_mstat,MEMORYSTATUS* out_mstat);
    void				__fastcall	DumpMemoryStatus(System::AnsiString descript,MEMORYSTATUS* in_mstat,MEMORYSTATUS* out_mstat);
    
public:
						__fastcall 	TDirectDraw(TComponent* Owner);
    virtual				__fastcall 	~TDirectDraw();

    HDC					__fastcall	GetBackBufferDC();
    HDC					__fastcall	GetPrimarySurfaceDC();
    bool				__fastcall	ReleaseBackBufferDC();
    bool				__fastcall	ReleasePrimarySurfaceDC();

	bool				__fastcall	Flip(bool wait = true);
    bool				__fastcall	FlipToGDI();
    bool				__fastcall	FlipGDIToBackBuffer(bool wait = true);
    bool				__fastcall	FlipBackBufferToGDI(bool wait = true);
    bool				__fastcall	FlipGDIRectToBackBuffer(int left,int top,int width,int height, bool wait = true);
    bool				__fastcall	FlipBackBufferRectToGDI(int left,int top,int width,int height, bool wait = true);

    bool				__fastcall	DrawMouseCursor(bool drawing_to_primary_surface = true);

    bool				__fastcall	BlitTo(HDC hdc,int x,int y,TImage* image,long copymode);
    bool				__fastcall	BlitToBackBuffer(int x,int y,TImage* image,long copymode = cmSrcCopy);
    bool				__fastcall	BlitToPrimarySurface(int x,int y,TImage* image,long copymode = cmSrcCopy);

    TImage*				__fastcall	GetImage(HDC hdc,int left,int top,int width,int height);
    TImage*				__fastcall	GetBackBufferImage(int left,int top,int width,int height);
    TImage*				__fastcall	GetPrimarySurfaceImage(int left,int top,int width,int height);

	bool				__fastcall	SetDisplayMode(int width,int height,int bitsperpixel);

    void				__fastcall	Log( int msgclass, System::AnsiString functiondesc, System::AnsiString msg );
    void				__fastcall	Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg );
    System::AnsiString	__fastcall	ErrorMessage(int errorvalue);

//	void				__fastcall	Clear(TColor color);

	__property	LPDIRECTDRAW		DirectDraw = {read=fDirectDraw};
    __property	HRESULT				ErrorValue = {read=fErrorValue};
    __property	int					ErrorCount = {read=fErrorCount,write=fErrorCount};
    __property	TStringList*		ErrorStrings = {read=fErrorStrings};

__published:

	__property bool					Debugging = {read=fDebugging,write=fDebugging,default=false};
    __property bool					Enabled = {read=fEnabled,write=SetEnabled,default=false};

	__property int					Width = {read=fWidth,write=SetWidth,default=800};
    __property int					Height = {read=fHeight,write=SetHeight,default=600};
    __property int					BitsPerPixel = {read=fBitsPerPixel,write=SetBitsPerPixel,default=16};
    __property int					BackSurfaceCount = {read=fBackBufferCount,write=SetBackSurfaceCount,default=1};
};
//---------------------------------------------------------------------------
extern  TDirectDraw*	DirectDraw;
//---------------------------------------------------------------------------
#endif

