//---------------------------------------------------------------------------
//
//	TDirectDrawForm - VCL & DirectX transparent implementation
//
//	(c) 1997 Darren J. Dwyer  - All Rights Reserved
//
// Disclaimer:
// All source code, object files, documentation files, etc. included
// with this package are copyright. Any use contrary to the attached
// file LICENCE.DOC are deemed to be breaches of copyright.
//
//---------------------------------------------------------------------------
#ifndef TDDFormH
#define TDDFormH
//---------------------------------------------------------------------------
#include <vcl\Classes.hpp>
#include <vcl\Controls.hpp>
#include <vcl\StdCtrls.hpp>
#include <vcl\Forms.hpp>
//---------------------------------------------------------------------------
#include <vcl\ExtCtrls.hpp>
#include <vcl\Buttons.hpp>
#include <vcl\ComCtrls.hpp>
#include "Grids.hpp"
//---------------------------------------------------------------------------
#include "TDirectDraw.h"
#include "dinput.h"
#include <vcl\MPlayer.hpp>
#include <vcl\Menus.hpp>
//---------------------------------------------------------------------------
class TDDForm : public TForm
{
__published:	// IDE-managed Components
	TPanel *TimingPanel;
	TLabel *FramesLabel;
	TLabel *SecondsLabel;
	TLabel *FramesPerSecondLabel;
	TPanel *InformationPanel;
	TListBox *DebugMessages;
	TBitBtn *ExitButton;
	TButton *ResetCounters;
	TTimer *AnimationTimer;
	TShape *Shape1;
	TBitBtn *RepaintMessagesButton;
	TBitBtn *ClearMessagesButton;
	TBitBtn *ToggleDebuggingButton;
	TLabel *AnimSpeedLabel;
	TLabel *TicksLabel;
	TButton *EnableDirectXButton;
	TLabel *DirectXInstalledLabel;
	TLabel *DirectXDebuggingLabel;
	TLabel *DirectXEnabledLabel;
	TImage *Image1;
	TBitBtn *BitBtn2;
	TBitBtn *BitBtn4;
	TRadioGroup *RadioGroup1;
	TMemo *Memo1;
	TEdit *Edit1;
	TScrollBar *ScrollBar1;
	TButton *Button1;
	TButton *Button2;
	TTimer *LittleAnimTimer;
	TImage *Image2;
	TScrollBar *ScrollBar2;
	TListBox *ListBox1;
	TSpeedButton *SpeedButton1;
	void __fastcall ExitButtonClick(TObject *Sender);
	void __fastcall EnableDirectXButtonClick(TObject *Sender);
	void __fastcall ResetCountersClick(TObject *Sender);
	void __fastcall AnimationTimerTimer(TObject *Sender);
	void __fastcall FormDeactivate(TObject *Sender);
	void __fastcall RepaintMessagesButtonClick(TObject *Sender);
	void __fastcall FormPaint(TObject *Sender);
	void __fastcall ToggleDebuggingButtonClick(TObject *Sender);
	void __fastcall ClearMessagesButtonClick(TObject *Sender);
	void __fastcall BitBtn2Click(TObject *Sender);
	void __fastcall BitBtn4Click(TObject *Sender);
	void __fastcall Button1Click(TObject *Sender);
	void __fastcall Button2Click(TObject *Sender);
	void __fastcall ScrollBar1Scroll(TObject *Sender, TScrollCode ScrollCode, int &ScrollPos);
	void __fastcall LittleAnimTimerTimer(TObject *Sender);
	void __fastcall FormCreate(TObject *Sender);
	void __fastcall SpeedButton1Click(TObject *Sender);
	
private:	// User declarations

	// BEGIN 			DirectXForm stuff
    bool				fHandlingWMPAINT;
    int					fMouseX;
    int					fMouseY;
    int					fLeft;
    int					fTop;
    int					fWidth;
    int					fHeight;
    bool				fFormIsActive;
    bool				fFormIsMoving;
    bool				fFormIsBeingRestored;
    bool				fFormIsBeingMinimized;
    bool				fFormIsBeingMaximized;
    bool				fMenuIsDisplayed;
	int					fBorderX;
	int					fBorderY;
	int					fCaptionHeight;
    int					fButtonWidth;
    int					fButtonHeight;
    int					fScrollbarWidth;
    int					fScrollbarHeight;
    int					fThumbnailHeight;
    int					fIconWidth;
    int					fIconHeight;
    bool				fExcludeMenuYes;
    TRect				fExcludeMenu;
    System::AnsiString	fCaptionFontName;

	// END 				DirectXForm stuff

	// other drawing stuff
	unsigned int		fAnimationTicks;
	unsigned int		fFramesFlipped;
    unsigned int		fSeconds;
    unsigned int		fColor;
    TDateTime 			flast_time;
	int 				fx;
	int 				fy;
	int 				fxincr;
	int 				fyincr;

protected:

		// BEGIN 	overloaded VCL functions used to draw the form / controls
		void 		__fastcall	WndProc(Messages::TMessage &Message);     	// Forms.Pas & Controls.Pas
		void 		__fastcall	PaintHandler(Messages::TWMPaint &Message);	// Controls.Pas
		void		__fastcall	PaintWindow(HDC hdc);						// Forms.Pas & Controls.Pas
		void		__fastcall	PaintControls(HDC hdc, TControl* First);	// Controls.Pas
		void		__fastcall	Repaint();
		// END 		overloaded VCL functions used to draw the form / controls

		// BEGIN extra functions used to draw the form / controls
        void		__fastcall	ExpandBorder( RECT* r, TBorderStyle borderstyle, bool include_caption );
        void		__fastcall	ReduceBorder( RECT* r, TBorderStyle borderstyle, bool include_caption );
        void		__fastcall	ExcludeOtherForms( HWND hwnd, HDC hdc );
        void		__fastcall	ExcludeMenu( HDC hdc, HMENU hmenu );
		void		__fastcall	PaintFrame( HDC hdc,RECT frame_r );
        void		__fastcall	PaintCommandButton( HDC hdc, int x, int y );
//        void		__fastcall	PaintCaption( HDC hdc, int x, int y );
        void		__fastcall	PaintMinimizeButton( HDC hdc, int x, int y );
        void		__fastcall	PaintMaximizeButton( HDC hdc, int x, int y );
        void		__fastcall	PaintRestoreButton( HDC hdc, int x, int y );
        void		__fastcall	PaintCloseButton( HDC hdc, int x, int y );
        void		__fastcall	MyFrame3DWindow( TCanvas* canvas, int left,int top,int right,int bottom,int penwidth );
		void		__fastcall	MyFrame3DButton( TCanvas* canvas, int left,int top,int right,int bottom);
		void		__fastcall	MyFrame3DThumbnail( TCanvas* canvas, int left, int top, int right, int bottom);
		void		__fastcall	PaintListBox( HDC hdc, TListBox* listbox );
        void		__fastcall	PaintVScrollBar( HDC hdc, HWND hwnd, int left, int top, int width, int height, int pos, int rowoffset, int rowsperpage, int totalrows );
        void		__fastcall	PaintHScrollBar( HDC hdc, HWND hwnd, int left, int top, int width, int height, int pos, int rowoffset, int rowsperpage, int totalrows );
        void		__fastcall	PaintControlVScrollBar( HDC hdc, int left, int top, int width, int height, int pos, int range );
        void		__fastcall	PaintControlHScrollBar( HDC hdc, int left, int top, int width, int height, int pos, int range );
		// END extra functions used to draw the form / controls

	void		__fastcall	PaintWithVCL();

public:		// User declarations
				__fastcall 	TDDForm(TComponent* Owner);
	virtual		__fastcall 	~TDDForm();
    void		__fastcall	UpdateDirectXMessages();
};
//---------------------------------------------------------------------------
extern TDDForm *DDForm;
//---------------------------------------------------------------------------
#endif


