//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TDirectDrawForm.h"
//---------------------------------------------------------------------------
#pragma link "TDirectDraw"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TDirectDrawTestForm *DirectDrawTestForm;
//---------------------------------------------------------------------------
__fastcall TDirectDrawTestForm::TDirectDrawTestForm(TComponent* Owner)
	: TForm(Owner)
{
fAnimationTicks = 0;
fFramesFlipped = 0;
fSeconds = 0;
fColor = 0;

fPaintWithVCL = false;
fPaintWithGDI = false;
fPaintWithDirectDrawDC = false;
fInsidePaintFunction = false;

fDrawingOnDirectXDC = false;
fDrawingOnDirectXDCWorked = false;
fSaveCanvasHandle = false;
fSaveCanvasHandleWorked = false;
fRestoreCanvasHandle = false;
fRestoreCanvasHandleWorked = false;
fDrawingVCLOnDirectX = false;
fDrawingVCLOnDirectXWorked = false;
fSurfaceWasFlipped = false;

fDirectDraw = new TDirectDraw(this);
if (fDirectDraw != NULL) DirectDraw->Debugging = false;
}
//---------------------------------------------------------------------------
__fastcall 	TDirectDrawTestForm::~TDirectDrawTestForm()
{
if (fDirectDraw != NULL) delete fDirectDraw;
}
//---------------------------------------------------------------------------
void 	__fastcall TDirectDrawTestForm::WndProc(Messages::TMessage &Message)
{
// handle any repaint messages coming from Windows rather than letting the default
// WndProc function handle them as the default WndProc function automatically draws
// the form when receiving WM_PAINT, WM_NCPAINT, WM_NCACTIVATE.....
// since the form redrawing code uses a windows gdi Paint3d function which blits to
// the display on completion (we have no control over this), it is probably better
// to use our own handler so that we can control exactly when to blit what.

HDC		hdc;

switch (Message.Msg)
	{
    case WM_PAINT:
//    case WM_NCPAINT:
//    case WM_NCACTIVATE:

		// draw background / basic form image
    	if (fDirectDraw != NULL)
            {
	        switch (Message.Msg)
	        	{
	        	case WM_PAINT: fDirectDraw->Log(LOG_INFO,"","HANDLING WM_PAINT"); break;
	            case WM_NCPAINT: fDirectDraw->Log(LOG_INFO,"","HANDLING WM_NCPAINT"); break;
	            case WM_NCACTIVATE: fDirectDraw->Log(LOG_INFO,"","HANDLING WM_NCACTIVATE"); break;
	            }

	        // this returns the backbuffer if possible, primary surface otherwise
	      	hdc = DirectDraw->GetDC();

	        if (hdc != NULL)
	           	{
	            fDrawingOnDirectXDC = true;
	            DirectDraw->Log(LOG_NOTE,"<<FORM>> WndProc()","DirectDraw->GetDC() was successful");

                // setup the canvas so all references to it use the DirectX supplied DC
                Canvas->Handle = hdc;

                // paint the form
                Canvas->Brush->Color = clGray;
                Canvas->Brush->Style = bsSolid;
                Canvas->Rectangle(Left,Top,Left+Width,Top+Height);

                // reset the canvas now that we've finished using it....
                Canvas->Handle = NULL;

                // perform normal draw functions (without painting the actual form)
                Messages::TWMPaint	MyPaintMessage;
                MyPaintMessage.Msg = WM_PAINT;
                MyPaintMessage.DC  = hdc;
                MyPaintMessage.Unused = 0;
                MyPaintMessage.Result = 0;
                PaintHandler(MyPaintMessage);

                // make sure to release the DirectX supplied DC
		        DirectDraw->Log(LOG_NOTE,"<<FORM>> WndProc()","just about to release the DC");
				DirectDraw->ReleaseDC(&hdc);

                // flip the DirectX buffers (thus making recent drawing visible).
		        DirectDraw->Log(LOG_NOTE,"<<FORM>> WndProc()","just about to flip the surface");
                for (int adelay=0;adelay<10000000;adelay++) ;
		        fSurfaceWasFlipped = DirectDraw->Flip();

		      	DirectDraw->Log(LOG_NOTE,"<<FORM>> WndProc()","!! Direct Drawing Worked !!");
	            }
	        else
		        TForm::WndProc(Message);
            }
        else
        	TForm::WndProc(Message);

        break;

    default:

    	TForm::WndProc(Message);
    }
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawTestForm::PaintControls(HDC hdc, TControl* First)
{
DirectDraw->Log(LOG_IN,"<<FORM>> PaintControls()","");
if (ControlCount == 0)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>> PaintControls()","ControlCount is 0");
    return;
    }

int i = 0;
if (hdc != NULL)
	{
	IntersectClipRect(hdc, 0, 0, Width, Height);
	int j=0;
    while (j<ControlCount)
    	{
		if (Controls[j] == First) i = j;
        j++;
        }
    }
Canvas->Handle = hdc;
int k = 0;
while (k<ControlCount)
	{
//    RECT r;
//    r.left = Controls[i]->Left;
//    r.top  = Controls[i]->Top;
//    r.right  = Controls[i]->Left + Controls[i]->Width;
//    r.bottom = Controls[i]->Top + Controls[i]->Height;

//   	if (Controls[i]->Visible) //&& (RectVisible(hdc, &r)))
    	{
		DirectDraw->Log(LOG_OUT,"<<FORM>> PaintControls()","Painting Control - '" + Controls[i]->Name + "'");

        MoveWindowOrg(hdc, Controls[i]->Left, Controls[i]->Top);
//        IntersectClipRect(hdc, 0, 0, Controls[i]->Width, Controls[i]->Height);

		// save current control style / state flags
        bool paintcopy = (Controls[i]->ControlState.Contains(csPaintCopy));

        // now make sure the control is replicatable and can be copied
        if (paintcopy) ControlState >> csPaintCopy;

//        Controls[i]->ControlStyle << csReplicatable;
//		Controls[i]->ControlState << csPaintCopy;
        Controls[i]->Perform(WM_ERASEBKGND, (long) hdc, NULL);
       	Controls[i]->Perform(WM_PAINT, (long) hdc, NULL);

        // restore the original style / state settings
        if (paintcopy) ControlState << csPaintCopy;
//		if (!is_paintcopy) Controls[i]->ControlState >> csPaintCopy;
//        if (!is_replicatable) Controls[i]->ControlStyle >> csReplicatable;

        MoveWindowOrg(hdc, -Controls[i]->Left, -Controls[i]->Top);
        }

    ++i %= ControlCount;	// wrap to 0..ControlCount-1
    k++;
    }
Canvas->Handle = NULL;
DirectDraw->Log(LOG_OUT,"<<FORM>> PaintControls()","painted " + IntToStr(k) + " controls");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::PaintHandler(Messages::TWMPaint &Message)
{
// flag that the draw has not yet worked
fDrawingOnDirectXDCWorked = false;

bool directdraw_created = (fDirectDraw != NULL);
if (directdraw_created && DirectDraw->Enabled)
   	{
	DirectDraw->Log(LOG_IN,"<<FORM>> PaintHandler()","DirectDraw is enabled & will be used for drawing");
    DirectDraw->Log(LOG_INFO,"<<FORM>> PaintHandler()","just about to get DirectDraw DC");

    // paint using standard VCL functions directly on the DirectX supplied surface
	fDrawingVCLOnDirectXWorked = false;
	DirectDraw->Log(LOG_INFO,"<<FORM>> PaintHandler()","about to VCL paint to the DirectDraw DC");

    // if the supplied DC is okay
    HDC hdc = Message.DC;
    if (hdc != NULL)
        // use standard function calls to paint to our DirectX surface
	    try
	    	{
		    DirectDraw->Log(LOG_NOTE,"<<FORM>> PaintHandler()","about to PaintWindow()");
	        PaintWindow(hdc);
			DirectDraw->Log(LOG_NOTE,"<<FORM>> PaintHandler()","about to PaintControls()");
		    PaintControls(hdc,NULL);

			fDrawingOnDirectXDCWorked = true;
			DirectDraw->Log(LOG_OUT,"<<FORM>> PaintHandler()","VCL painting on DirectDraw DC was successful");
	        }
        // unless a problem occurs
		catch (...)
	       	{
	        fDrawingOnDirectXDCWorked = false;
	        DirectDraw->Log(LOG_OUT,"<<FORM>> PaintHandler()","VCL painting on DirectDraw DC NOT successful");
	        }
    }

// if the direct drawing did not work, call the standard PaintHandler()
if (!fDrawingOnDirectXDCWorked)
	{
    Message.DC = 0;
    TForm::PaintHandler(Message);
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::ExitButtonClick(TObject *Sender)
{
if (Enabled) Close();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::PaintWithVCLButtonClick(TObject *Sender)
{
fPaintWithVCL = !fPaintWithVCL;
fPaintWithGDI = false;
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::PaintWithGDIButtonClick(TObject *Sender)
{
fPaintWithGDI = !fPaintWithGDI;
fPaintWithVCL = false;
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::PaintWithDirectDrawButtonClick(TObject *Sender)
{
fPaintWithDirectDrawDC = !fPaintWithDirectDrawDC;
DirectDraw->Enabled = fPaintWithDirectDrawDC;
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::UpdateDirectXMessages()
{
if (DirectDraw->ErrorStrings)
	{
	DirectDrawErrorMessagesBox->Items->Clear();
	for (int i=0;i<DirectDraw->ErrorStrings->Count;i++)
		DirectDrawErrorMessagesBox->Items->Add(DirectDraw->ErrorStrings->Strings[i]);
	DirectDraw->ErrorCount = 0;
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::FormCreate(TObject *Sender)
{
bool all_enabled = true;
DirectDraw->Enabled = true;
if (DirectDraw->Enabled)	
	{
	if (DirectDraw->SetDisplayMode(640,480,8))
    	{
	    AnimationTimer->Interval = AutoFlipTimer->Interval;
	    AnimationTimer->Enabled = true;
	    AutoFlipTimer->Enabled = true;
        all_enabled = true;
        }
    }
//else
//	ShowMessage("DirectDraw could not be enabled");

if (!all_enabled)
	{
    for (int i=0;i<ControlCount;i++) Controls[i]->Enabled = false;
    AnimationTimer->Enabled = false;
    AutoFlipTimer->Enabled = false;
    }

Left=0; Width=DirectDraw->Width-Left;
Top=5; Height=DirectDraw->Height-Top;
DirectDrawErrorMessagesBox->Height -= 5;

fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::FormActivate(TObject *Sender)
{
AnimationTimer->Enabled = true;
if (DirectDraw->ErrorValue == DD_OK) DirectXInstalledLabel->Caption = "DirectX is Installed";
else DirectXInstalledLabel->Caption = "DirectX is Not Installed";
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::FormDeactivate(TObject *Sender)
{
AnimationTimer->Enabled = false;
AutoFlipTimer->Enabled = false;
DirectDraw->Enabled = false;
DirectDrawErrorMessagesBox->Items->Clear();
}
//---------------------------------------------------------------------------
int PageNumber;
void __fastcall TDirectDrawTestForm::FormPaint(TObject *Sender)
{
++PageNumber %= 2;	// 0..1 only

Canvas->Pen->Color = clBlack;
Canvas->Brush->Style = bsSolid;
Canvas->Rectangle(Left,Top,Left+Width,Top+Height);

Canvas->Pen->Color = clRed;
Canvas->Font->Color = clWhite;
Canvas->TextOut(0,0,"PAGE # " + IntToStr(PageNumber));

int seconds_elapsed = fSeconds;

// BEGIN REGULAR DRAW CODE
TicksLabel->Caption = "Ticks: " + IntToStr(fAnimationTicks);
FramesLabel->Caption = "Frames: " + IntToStr(fFramesFlipped);
SecondsLabel->Caption = "Seconds: " + IntToStr(seconds_elapsed);
if (seconds_elapsed != 0)
	{
    FramesPerSecondLabel->Caption = "FPS: " + FloatToStrF((float)fFramesFlipped / (float)seconds_elapsed, ffFixed, 6, 2);
	TicksPerSecondLabel->Caption = "TPS: " + FloatToStrF((float)fAnimationTicks / (float)seconds_elapsed, ffFixed, 6, 2);
	}

if (fPaintWithVCL)
	{
	if (fPaintWithDirectDrawDC)
    	{
		PaintWithVCLButton->Caption = "Painting with &VCL on DIRECTX surface: ON";
        PaintWithLabel->Caption = "Using standard VCL functions to paint on a DirectX surface";
		PaintWithLabel->Font->Color = clGreen;
        }
    else
    	{
        PaintWithVCLButton->Caption = "Paint with &VCL on VCL canvas: ON";
        PaintWithLabel->Caption = "Using standard VCL functions to paint on a VCL canvas";
		PaintWithLabel->Font->Color = clRed;
        }
    PaintWithVCL();
    }
else
	{
    if (fPaintWithDirectDrawDC)	PaintWithVCLButton->Caption = "Paint with &VCL on DIRECTX surface: OFF";
	else PaintWithVCLButton->Caption = "Paint with &VCL on VCL canvas: OFF";
    }
if (fPaintWithGDI)
	{
	if (fPaintWithDirectDrawDC)
    	{
        PaintWithGDIButton->Caption = "Paint with &GDI on DIRECTX surface: ON";
        PaintWithLabel->Caption = "Using standard GDI functions to paint on a DirectX surface";
		PaintWithLabel->Font->Color = clGreen;
        }
    else
    	{
        PaintWithGDIButton->Caption = "Paint with &GDI on VCL canvas: ON";
        PaintWithLabel->Caption = "Using standard GDI functions to paint on a VCL canvas";
		PaintWithLabel->Font->Color = clRed;
        }
    PaintWithGDI();
    }
else
	{
	if (fPaintWithDirectDrawDC) PaintWithGDIButton->Caption = "Paint with &GDI on DIRECTX surface: OFF";
    else PaintWithGDIButton->Caption = "Paint with &GDI on VCL canvas: OFF";
    }
if (fPaintWithDirectDrawDC)PaintWithDirectDrawButton->Caption = "Paint directly on &DirectX surface: ON";
else PaintWithDirectDrawButton->Caption = "Paint directly on &DirectX surface: OFF";

if (DirectDraw->Debugging)
	{
    DebuggingLabel->Font->Color = clGreen;
    DebuggingLabel->Caption = "DirectX debugging is ON";
    }
else
	{
    DebuggingLabel->Font->Color = clRed;
    DebuggingLabel->Caption = "DirectX debugging is OFF";
    }

if (fDrawingOnDirectXDC)
	{
    DrawingOnDirectXDCLabel->Font->Color = clGreen;
    DrawingOnDirectXDCLabel->Caption = "drawing on DirectX supplied surface";
    }
else
	{
    DrawingOnDirectXDCLabel->Font->Color = clRed;
    DrawingOnDirectXDCLabel->Caption = "drawing on VCL supplied surface";
    }

if (fDrawingOnDirectXDCWorked)
	{
    DrawingOnDirectXDCWorkedLabel->Font->Color = clGreen;
    DrawingOnDirectXDCWorkedLabel->Caption = "drawing on DirectX surface worked";
    }
else
	{
    DrawingOnDirectXDCWorkedLabel->Font->Color = clRed;
    DrawingOnDirectXDCWorkedLabel->Caption = "drawing on DirectX surface failed";
    }

if (fSaveCanvasHandle)
	{
    SaveCanvasHandleLabel->Font->Color = clGreen;
    SaveCanvasHandleLabel->Caption = "canvas handle was saved (freed)";
    }
else
	{
    SaveCanvasHandleLabel->Font->Color = clRed;
    SaveCanvasHandleLabel->Caption = "canvas handle was not saved (freed)";
    }

if (fSaveCanvasHandleWorked)
	{
    SaveCanvasHandleWorkedLabel->Font->Color = clGreen;
    SaveCanvasHandleWorkedLabel->Caption = "canvas handle was saved (freed) correctly";
    }
else
	{
    SaveCanvasHandleWorkedLabel->Font->Color = clRed;
    SaveCanvasHandleWorkedLabel->Caption = "canvas handle did not save (free) correctly";
    }

if (fRestoreCanvasHandle)
	{
    RestoreCanvasHandleLabel->Font->Color = clGreen;
    RestoreCanvasHandleLabel->Caption = "canvas handle was restored";
    }
else
	{
    RestoreCanvasHandleLabel->Font->Color = clRed;
    RestoreCanvasHandleLabel->Caption = "canvas handle was not not restored";
    }

if (fRestoreCanvasHandleWorked)
	{
    RestoreCanvasHandleWorkedLabel->Font->Color = clGreen;
    RestoreCanvasHandleWorkedLabel->Caption = "canvas handle was restored (zeroed) correctly";
    }
else
	{
    RestoreCanvasHandleWorkedLabel->Font->Color = clRed;
    RestoreCanvasHandleWorkedLabel->Caption = "canvas handle did not restore (zero) correctly";
    }

if (fDrawingVCLOnDirectX)
	{
    DrawingVCLOnDirectXLabel->Caption = "attempting to use VCL to draw on DirectX DC";
    }
else
	{
    DrawingVCLOnDirectXLabel->Caption = "not using VCL to draw on DirectX DC ?? reason?";
    }

if (fDrawingVCLOnDirectXWorked)
	{
    DrawingVCLOnDirectXWorkedLabel->Caption = "Drawing using VCL on DirectX surface worked";
    }
else
	{
    DrawingVCLOnDirectXWorkedLabel->Caption = "Drawing using VCL on DirectX not working";
    }

if (fSurfaceWasFlipped)
	{
    
    SurfaceWasFlippedLabel->Caption = "surface was flipped";
    }
else
	{
    SurfaceWasFlippedLabel->Caption = "surface was not flipped";
    }
}
//---------------------------------------------------------------------------
int x = 0;
int y = 0;
int xincr = 1;
int yincr = 1;
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawTestForm::PaintWithVCL()
{
// animation code
fColor++;
x += xincr;
y += yincr;

if (x<0) xincr = 1;
if (y<0) yincr = 1;
if (x+400>Width) xincr = -1;
if (y+200>Height) yincr = -1;

Canvas->Font->Color = clWhite;
Canvas->TextOut(700,0,"X == " + IntToStr(x));

// use the standard VCL functions to draw on the DirectX surface as we see fit
Canvas->Pen->Color = (TColor) RGB(255,255,255);
Canvas->Brush->Color = (TColor) RGB((x*128)%256,(x*16)%256,(x*16+16*fColor)%256) ;
Canvas->Brush->Style = bsSolid;
Canvas->Rectangle( x,y, x+400,y+200 );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawTestForm::PaintWithGDI()
{
// use standard GDI functions to manipulate the DirectX surface as we see fit
SetTextColor( Canvas->Handle, RGB(255,255,255));
SetBkColor( Canvas->Handle, RGB((x*128)%256,(x*16)%256,(x*16+16*fColor)%256));
RECT r;
r.left = x;
r.top = y;
r.right = x+400;
r.bottom = y+200;
FillRect( Canvas->Handle, &r, 0 );
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::BitBtn1Click(TObject *Sender)
{
if (Enabled)
	{
	UpdateDirectXMessages();
	DirectDrawErrorMessagesBox->Refresh();
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::BitBtn2Click(TObject *Sender)
{
if (Enabled)
	{
    DirectDraw->ErrorStrings->Clear();
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::BitBtn3Click(TObject *Sender)
{
if (Enabled)
	{
    DirectDraw->Debugging = !DirectDraw->Debugging;
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::SecondsTimerTimer(TObject *Sender)
{
fSeconds++;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::AnimationTimerTimer(TObject *Sender)
{
fAnimationTicks++;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::AutoFlipTimerTimer(TObject *Sender)
{
fFramesFlipped++;
Refresh();
}
//---------------------------------------------------------------------------
// remember: function SetPriorityClass(hProcess: THandle; dwPriorityClass: DWORD): BOOL; stdcall;

void __fastcall TDirectDrawTestForm::BitBtn4Click(TObject *Sender)
{
if (Enabled) if (AnimationTimer->Interval>20) AnimationTimer->Interval -= 20;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::BitBtn5Click(TObject *Sender)
{
if (Enabled) if (AnimationTimer->Interval<2000) AnimationTimer->Interval += 20;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawTestForm::FormDestroy(TObject *Sender)
{
DirectDrawErrorMessagesBox->Items->Clear();
}
//---------------------------------------------------------------------------
