//---------------------------------------------------------------------------
#ifndef TDirectDrawH
#define TDirectDrawH
//---------------------------------------------------------------------------
#include <vcl\SysUtils.hpp>
#include <vcl\Controls.hpp>
#include <vcl\Classes.hpp>
#include <vcl\Forms.hpp>
//---------------------------------------------------------------------------
#include <d3drmwin.h>
#include <direct.h>
//---------------------------------------------------------------------------
#define	LOG_IN			0
#define	LOG_OUT			1
#define	LOG_INFO		2
#define	LOG_NOTE		3
#define	LOG_WARNING		4
#define	LOG_ERROR		5
//---------------------------------------------------------------------------
class TDirectDraw : public TComponent
{
private:
    HRESULT							fErrorValue;
    int								fErrorCount;
    TStringList*					fErrorStrings;

	LPDIRECTDRAW					fDirectDraw;
    LPDIRECTDRAWSURFACE				fPrimarySurface;
    LPDIRECTDRAWSURFACE				fBackBuffer;
//	DDSURFACEDESC 					fBackBufferLockingDesc;
//	DDSURFACEDESC 					fPrimarySurfaceLockingDesc;

    HDC								fHandle;
    int								fWidth;
    int								fHeight;
    int								fBitsPerPixel;
    int								fBackBufferCount;

    bool							fExclusiveModeSet;
    bool							fDisplayModeChanged;

    bool							fDebugging;
    bool							fEnabled;

    bool							fBackBufferDCRetrieved;
    bool							fPrimarySurfaceDCRetrieved;

    // functions used internally
    bool				__fastcall	CreateDirectDraw();
    bool				__fastcall	ReleaseDirectDraw();
    bool				__fastcall	SetExclusiveMode();
    bool				__fastcall	SetNormalMode();
    bool				__fastcall	RestoreDisplayMode();
    bool				__fastcall	FlipToGDISurface();
	bool				__fastcall	CreateSurfaces();
    bool				__fastcall	ReleaseSurfaces();

    // property functions
	void				__fastcall	SetWidth(int width);
    void				__fastcall	SetHeight(int height);
	void				__fastcall	SetBitsPerPixel(int bitsperpixel);
	void				__fastcall	SetBackSurfaceCount(int backsurfacecount);

protected:
public:
						__fastcall 	TDirectDraw(TComponent* Owner);
    virtual				__fastcall 	~TDirectDraw();

    HDC					__fastcall	GetDC(bool primaryrequired = false);
    bool				__fastcall	ReleaseDC(HDC& hdc);
	bool				__fastcall	Flip(bool wait = true);

	bool				__fastcall	SetDisplayMode(int width,int height,int bitsperpixel);

    void				__fastcall	Log( int msgclass, System::AnsiString functiondesc, System::AnsiString msg );
    void				__fastcall	Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg );
    System::AnsiString	__fastcall	ErrorMessage(int errorvalue);

//	void				__fastcall	Clear(TColor color);

	__property	LPDIRECTDRAW		DirectDraw = {read=fDirectDraw};
    __property	HDC					Handle = {read=fHandle};
    __property	HRESULT				ErrorValue = {read=fErrorValue};
    __property	int					ErrorCount = {read=fErrorCount,write=fErrorCount};
    __property	TStringList*		ErrorStrings = {read=fErrorStrings};

__published:

	__property bool					Debugging = {read=fDebugging,write=fDebugging,default=false};
    __property bool					Enabled = {read=fEnabled,write=fEnabled,default=false};

	__property int					Width = {read=fWidth,write=SetWidth,default=800};
    __property int					Height = {read=fHeight,write=SetHeight,default=600};
    __property int					BitsPerPixel = {read=fBitsPerPixel,write=SetBitsPerPixel,default=16};
    __property int					BackSurfaceCount = {read=fBackBufferCount,write=SetBackSurfaceCount,default=1};
};
//---------------------------------------------------------------------------
#endif

