//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <fcntl.h>
#include <io.h>
#include <sys\stat.h>
//---------------------------------------------------------------------------
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TDirectDraw.h"
//---------------------------------------------------------------------------
static inline TDirectDraw *ValidCtrCheck()
{
	return new TDirectDraw(NULL);
}
//---------------------------------------------------------------------------
namespace Tdirectdraw
{
	void __fastcall Register()
	{
		TComponentClass classes[1] = {__classid(TDirectDraw)};
		RegisterComponents("DirectDraw", classes, 0);
	}
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::TDirectDraw(TComponent* Owner)
	: TComponent(Owner)
{
// create things used for debugging
fErrorValue = 0;
fErrorCount = 0;
fErrorStrings = new TStringList();
fErrorStrings->Sorted = false;
fErrorStrings->Duplicates = dupAccept;
fDebugging = true;
Log(LOG_IN,"TDirectDraw::Constructor()","Owner = >>> " + Owner->Name + " <<<");

// initialize some values
Log(LOG_INFO,"TDirectDraw::Constructor()","about to initialize internals");
fDirectDraw = NULL;
fPrimarySurface = NULL;
fBackBuffer = NULL;

fHandle = NULL;
fWidth = 800;
fHeight = 600;
fBitsPerPixel = 16;
fBackBufferCount = 1;

fEnabled = false;
fExclusiveModeSet = false;
fDisplayModeChanged = false;
fBackBufferDCRetrieved = false;
fPrimarySurfaceDCRetrieved = false;

// if we have created our DirectDraw object...
if (CreateDirectDraw())
	{
	// make a note of a successful result
	Log(LOG_NOTE,"TDirectDraw::Constructor()","the DirectDraw object has been created successfully");

	// try setting exclusive mode
    if (SetExclusiveMode())
    	{
	    // make a note of a successful result
		Log(LOG_NOTE,"TDirectDraw::Constructor()","exclusive mode has been set successfully");
        }

// temporarily - this works fine
//DirectDraw->SetDisplayMode(640,480,8);

    // create required display surfaces
    if (CreateSurfaces())
    	{
	    // make a note of a successful result
		Log(LOG_NOTE,"TDirectDraw::Constructor()","surfaces have been created successfully");
        }
    }
// otherwise display an error messsage
else
    ShowMessage("ERROR>> TDirectDraw::Constructor() >> Could not create the DirectDraw object");

Log(LOG_OUT,"TDirectDraw::Constructor()",">>> The Direct Draw object has been created successfully <<<");
fDebugging = false;
}
//---------------------------------------------------------------------------
__fastcall TDirectDraw::~TDirectDraw()
{
Log(LOG_IN,"TDirectDraw::Destructor()","");

// if we previously created our DirectDraw object...
if (fDirectDraw != NULL)
	{
	// make a note of a successful result
	Log(LOG_NOTE,"TDirectDraw::Destructor()","DirectDraw object was created, time to release");

    // release the display surfaces
    ReleaseSurfaces();

    // restore the original video mode
    RestoreDisplayMode();

	// restore normal mode
    SetNormalMode();

    // release the directdraw object
    ReleaseDirectDraw();
    }
// otherwise display an error messsage
else
    ShowMessage("ERROR>> TDirectDraw::Constructor() >> Could not release the DirectDraw object");

// release things used for debugging
if (fErrorStrings != NULL) { fErrorStrings->Clear(); delete fErrorStrings; }
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::CreateDirectDraw()
{
// simply call the DirectDrawCreate function & use result
fErrorValue = DirectDrawCreate(NULL, &fDirectDraw, NULL);
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateDirectDraw()","DirectDraw object could not be created");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::ReleaseDirectDraw()
{
if (fDirectDraw == NULL) return false;
// simply call the Release function & use result
fErrorValue = fDirectDraw->Release();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseDirectDraw()","DirectDraw object could not be released");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetExclusiveMode()
{
if (fDirectDraw == NULL) return false;
// simply call the SetCooperativeLevel function & use result
fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetExclusiveMode()","exclusive mode could not be set");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetNormalMode()
{
if (fDirectDraw == NULL) return false;
// simply call the SetCooperativeLevel function & use result
fErrorValue = fDirectDraw->SetCooperativeLevel( Application->Handle, DDSCL_NORMAL );
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetNormalMode()","normal mode could not be set");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::SetEnabled(bool enabled)
{
FlipToGDI();
fEnabled = enabled;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::RestoreDisplayMode()
{
if (fDirectDraw == NULL) return false;
// simply call the RestoreDisplayMode function & use result
fErrorValue = fDirectDraw->RestoreDisplayMode();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::RestoreVideoMode()","video mode could not be restored");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::CreateSurfaces()
{
if (fDirectDraw == NULL) return false;

// create the primary surface if it's not already created
fErrorValue = DD_OK;
if (fPrimarySurface == NULL)
	{
	// define the properties of the primary surface to be created
	DDSURFACEDESC	ddSurfaceDesc;
	memset(&ddSurfaceDesc,0,sizeof(ddSurfaceDesc));

	ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);
	ddSurfaceDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
	ddSurfaceDesc.dwBackBufferCount = fBackBufferCount;

	// create the primary surface
	fErrorValue = fDirectDraw->CreateSurface(&ddSurfaceDesc,&fPrimarySurface,NULL);
	if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not create the primary surface");
    else
    	{
        // note a successful result
        Log(LOG_NOTE,"TDirectDraw::CreateSurfaces()","the primary surface has been created");

        // only retrieve back buffer if any exist
        if (fBackBufferCount>0)
        	{
			// setup the surface caps ready to retrieve the back buffer
			DDSCAPS			ddsCaps;
			ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;

	        // retrieve a pointer to the 1st back buffer in the list (i theorize)
			fErrorValue = fPrimarySurface->GetAttachedSurface(&ddsCaps,&fBackBuffer);
			if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::CreateSurfaces()","could not retrieve back buffer");
		    else Log(LOG_NOTE,"TDirectDraw::CreateSurfaces()","back buffer has been retrieved");
            }
        }
	}
else
	Log(LOG_WARNING,"TDirectDraw::CreateSurfaces()","primary surface has already been created");

return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::ReleaseSurfaces()
{
if (fDirectDraw == NULL) return false;
/*
NOTE:	the back buffer cannot be released directly - all associated back
		buffers are released when you release the primary surface			*/

// get rid of our primary surface & any attached back buffers if they exist
fErrorValue = DD_OK;
if (fPrimarySurface != NULL)
	{
    fErrorValue = fPrimarySurface->Release();
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseSurfaces()","could not release primary surface");
	else
    	{
        fPrimarySurface = NULL;
    	fBackBuffer = NULL;
        }
    }
else
	Log(LOG_WARNING,"TDirectDraw::ReleaseSurfaces()","Primary surface was not allocated, thus no buffers were released");
return (fErrorValue != DD_OK);
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetWidth(int width)
{
if (!fEnabled) return;

// simply try to set the new display mode, if it works the width will be set
SetDisplayMode(width,fHeight,fBitsPerPixel);
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetHeight(int height)
{
if (!fEnabled) return;

// simply try to set the new display mode, if it works the height will be set
SetDisplayMode(fWidth,height,fBitsPerPixel);
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetBitsPerPixel(int bitsperpixel)
{
if (!fEnabled) return;

// simply try to set the new display mode, if it works the bitsperpixel will be set
SetDisplayMode(fWidth,fHeight,bitsperpixel);
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDraw::SetBackSurfaceCount(int backsurfacecount)
{
if (!fEnabled) return;

// only change the number of surfaces when required
if (backsurfacecount != fBackBufferCount)
	{
    // blow away any current surfaces,
    ReleaseSurfaces();
    // set the new buffer count,
    fBackBufferCount = backsurfacecount;
    // recreate the surfaces with the new number of buffers
    CreateSurfaces();
	}
}
//---------------------------------------------------------------------------
MEMORYSTATUS dc_in_mstat;
MEMORYSTATUS dc_out_mstat;
HDC	__fastcall TDirectDraw::GetDC(bool primaryrequired)
{
if (fDirectDraw == NULL) return NULL;
if (!fEnabled) return NULL;

// if a DC has been retrieved for either the primary surface or the backbuffer, fail
if (fBackBufferDCRetrieved || fPrimarySurfaceDCRetrieved) return NULL;

InMemoryStatus(&dc_in_mstat);

// setup some bits & pieces
HDC hdc;
fHandle = NULL;
fBackBufferDCRetrieved = false;
fPrimarySurfaceDCRetrieved = false;

// by default, and if the back buffer exists, try retrieving a DC for the back buffer
if ((fBackBuffer != NULL) && (!primaryrequired))
   	{
	fErrorValue = fBackBuffer->GetDC(&hdc);
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::GetDC()","Could not retrieve a DC for the back buffer");
    else
       	{
		fHandle = hdc;
	    fBackBufferDCRetrieved = true;
		Log(LOG_NOTE,"TDirectDraw::GetDC()","a DC for the back buffer was retrieved");
        }
	}
else
	if (!primaryrequired) Log(LOG_WARNING,"TDirectDraw::GetDC()","could not retrieve DC for back buffer cause it doesn't exist");
	else Log(LOG_WARNING,"TDirectDraw::GetDC()","could not retrieve DC for back buffer cause not requested");

// if neither DC has already been retrieved, try getting a DC for the primary surface
if (!fBackBufferDCRetrieved)
	if (fPrimarySurface != NULL)
	   	{
		fErrorValue = fPrimarySurface->GetDC(&hdc);
	    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::GetDC()","Could not retrieve a DC for the primary surface");
	    else
    	   	{
			fHandle = hdc;
	    	fPrimarySurfaceDCRetrieved = true;
			Log(LOG_NOTE,"TDirectDraw::GetDC()","a DC for the primary surface was retrieved");
	        }
		}
	else
		Log(LOG_WARNING,"TDirectDraw::GetDC()","could not retrieve DC for primary surface cause it doesn't exist");

// if a DC was retrieved for either surface, return the DC
return fHandle;
}
//---------------------------------------------------------------------------
bool	__fastcall TDirectDraw::ReleaseDC(HDC hdc)
{
if (fDirectDraw == NULL) return false;
if (!fEnabled) return false;
if (hdc == NULL) return false;

// if neither the backbuffer or the primary surface DC's were retrieved
if (!(fBackBufferDCRetrieved || fPrimarySurfaceDCRetrieved)) return false;

// if we retrieved the back buffer DC (using GetDC()), release it all now
if (fBackBufferDCRetrieved)
	{
	fErrorValue = fBackBuffer->ReleaseDC(hdc);
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseDC()","Could not release the back buffer DC");
	else
    	{
        fHandle = NULL;
        fBackBufferDCRetrieved = false;
        Log(LOG_NOTE,"TDirectDraw::ReleaseDC()","the back buffer DC has been released");
        }
    }
// if the back buffer was not retrieved, it must be the primary surface
else
	// if we retrieved the primary surface DC (using GetDC()), release it all now
	if (fPrimarySurfaceDCRetrieved)
		{
		fErrorValue = fPrimarySurface->ReleaseDC(hdc);
	    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::ReleaseDC()","Could not release the primary surface DC");
		else
	    	{
	        fHandle = NULL;
	        fPrimarySurfaceDCRetrieved = false;
	        Log(LOG_NOTE,"TDirectDraw::ReleaseDC()","the primary surface DC has been released");
	        }
	    }
OutMemoryStatus(&dc_out_mstat);
if (!SameMemoryStatus(&dc_in_mstat,&dc_out_mstat)) DumpMemoryStatus("GetDC/ReleaseDC MEMORY STATUS", &dc_in_mstat,&dc_out_mstat);
return (fHandle == NULL);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::Flip(bool wait)
{
if (fDirectDraw == NULL) return false;
if (!fEnabled) return false;

//
// NOTE: only surfaces with DDSCAPS_FLIP and DDSCAPS_FRONTBUFFER can be flipped
//		 ie. the primary surface
//

fErrorValue = DD_OK + 1;
if (fPrimarySurface != NULL)
   	{
    int flags = 0;
    if (wait) flags |= DDFLIP_WAIT;
    fErrorValue = fPrimarySurface->Flip(NULL,flags);
    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::Flip()","the primary surface could not be flipped");
    else Log(LOG_NOTE,"TDirectDraw::Flip()","the primary surface has been flipped successfully");
//    for (int adelay=0;adelay<250000000;adelay++) ;
    }
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::FlipToGDI()
{
if (fDirectDraw == NULL) return false;

// simply call the FlipToGDISurface function & use result
fErrorValue = fDirectDraw->FlipToGDISurface();
if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::FlipToGDI()","couldn't flip to gdi surface");
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SetDisplayMode(int width,int height,int bitsperpixel)
{
if (fDirectDraw == NULL) return false;
if (!fEnabled) return false;

// since the display mode can only be changed when in "Exclusive Mode"...
fDisplayModeChanged = false;
fErrorValue = DD_OK;
if (fExclusiveModeSet)
	{
    // make sure to release the surfaces & change the mode if and only if required
    if ((width != fWidth) || (height != fHeight) || (bitsperpixel != fBitsPerPixel))
    	{
        // tell DirectDraw to change the display mode to that required
	    fErrorValue = fDirectDraw->SetDisplayMode(width,height,bitsperpixel);
	    if (fErrorValue != DD_OK) Error(fErrorValue,"TDirectDraw::SetDisplayMode()","could not set display mode to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
	    else
	    	{
		    fWidth = width;
		    fHeight = height;
		    fBitsPerPixel = bitsperpixel;
		    fDisplayModeChanged = true;
			Log(LOG_NOTE,"TDirectDraw::SetDisplayMode()","the display mode has been changed to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
	        }
        }
    else
    	{
    	// since these are the current display mode settings,
        // return true as if the display mode was changed
    	fErrorValue = DD_OK;
		Log(LOG_INFO,"TDirectDraw::SetDisplayMode()","display mode was already set to: " + IntToStr(width) + "x" + IntToStr(height) + "x" + IntToStr(bitsperpixel));
        }
    }
return (fErrorValue == DD_OK);
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::InMemoryStatus(MEMORYSTATUS* in_mstat)
{
in_mstat->dwLength = sizeof(MEMORYSTATUS);
GlobalMemoryStatus( in_mstat );
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::OutMemoryStatus(MEMORYSTATUS* out_mstat)
{
out_mstat->dwLength = sizeof(MEMORYSTATUS);
GlobalMemoryStatus( out_mstat );
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::SameMemoryStatus(MEMORYSTATUS* in_mstat,MEMORYSTATUS* out_mstat)
{
return ((in_mstat->dwMemoryLoad == out_mstat->dwMemoryLoad) &&
		(in_mstat->dwTotalPhys == out_mstat->dwTotalPhys) &&
        (in_mstat->dwAvailPhys == out_mstat->dwAvailPhys) &&
        (in_mstat->dwTotalPageFile == out_mstat->dwTotalPageFile) &&
        (in_mstat->dwAvailPageFile == out_mstat->dwAvailPageFile) &&
        (in_mstat->dwTotalVirtual == out_mstat->dwTotalVirtual) &&
        (in_mstat->dwAvailVirtual == out_mstat->dwAvailVirtual));
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::DumpMemoryStatus(System::AnsiString descript,MEMORYSTATUS* in_mstat,MEMORYSTATUS* out_mstat)
{
int dwMemoryLoad = out_mstat->dwMemoryLoad-in_mstat->dwMemoryLoad;
int dwTotalPhys = out_mstat->dwTotalPhys-in_mstat->dwTotalPhys;
int dwAvailPhys = out_mstat->dwAvailPhys-in_mstat->dwAvailPhys;
int dwTotalPageFile = out_mstat->dwTotalPageFile-in_mstat->dwTotalPageFile;
int dwAvailPageFile = out_mstat->dwAvailPageFile-in_mstat->dwAvailPageFile;
int dwTotalVirtual = out_mstat->dwTotalVirtual-in_mstat->dwTotalVirtual;
int dwAvailVirtual = out_mstat->dwAvailVirtual-in_mstat->dwAvailVirtual;

Log(LOG_MEMSTATUS,"",">>>> " + descript + " <<<<");
if (dwMemoryLoad < 0) Log(LOG_MEMSTATUS,"","Memory Load: -" + IntToStr(dwMemoryLoad));
else if (dwMemoryLoad > 0) Log(LOG_MEMSTATUS,"","Memory Load: +" + IntToStr(dwMemoryLoad));

if (dwTotalPhys < 0) Log(LOG_MEMSTATUS,"","Total Phys: -" + IntToStr(dwTotalPhys));
else if (dwTotalPhys > 0) Log(LOG_MEMSTATUS,"","Total Phys: +" + IntToStr(dwTotalPhys));

if (dwAvailPhys < 0) Log(LOG_MEMSTATUS,"","Avail Phys: -" + IntToStr(dwAvailPhys));
else if (dwAvailPhys > 0) Log(LOG_MEMSTATUS,"","Avail Phys: +" + IntToStr(dwAvailPhys));

if (dwTotalPageFile < 0) Log(LOG_MEMSTATUS,"","Total Page: -" + IntToStr(dwTotalPageFile));
else if (dwTotalPageFile > 0) Log(LOG_MEMSTATUS,"","Total Page: +" + IntToStr(dwTotalPageFile));

if (dwAvailPageFile < 0) Log(LOG_MEMSTATUS,"","Avail Page: -" + IntToStr(dwAvailPageFile));
else if (dwAvailPageFile > 0) Log(LOG_MEMSTATUS,"","Avail Page: +" + IntToStr(dwAvailPageFile));

if (dwTotalVirtual < 0) Log(LOG_MEMSTATUS,"","Total Virt: -" + IntToStr(dwTotalVirtual));
else if (dwTotalVirtual > 0) Log(LOG_MEMSTATUS,"","Total Virt: +" + IntToStr(dwTotalVirtual));

if (dwAvailVirtual < 0) Log(LOG_MEMSTATUS,"","Avail Virt: -" + IntToStr(dwAvailVirtual));
else if (dwAvailVirtual > 0) Log(LOG_MEMSTATUS,"","Avail Virt: +" + IntToStr(dwAvailVirtual));
}
//---------------------------------------------------------------------------
MEMORYSTATUS	overall_in_mstat;
MEMORYSTATUS	overall_out_mstat;
bool			overall_FirstTimeIn = true;
MEMORYSTATUS	log_in_mstat;
MEMORYSTATUS	log_out_mstat;
int				log_indent = 0;
MEMORYSTATUS	log_before_mstat;
MEMORYSTATUS	log_after_mstat;
void	__fastcall	TDirectDraw::Log( int msgclass, System::AnsiString functiondesc, System::AnsiString msg )
{
if (!fDebugging) return;
TStringList*	new_strings = new TStringList();
if (new_strings == NULL) return;	// blah

// keep track of overall memory usage
//if (overall_FirstTimeIn)
//	{
//    overall_FirstTimeIn = false;
//    InMemoryStatus( &overall_in_mstat );
//    }
//if ((msgclass == LOG_OUT) || (msgclass == LOG_IN))
//	{
//	OutMemoryStatus( &overall_out_mstat );
//	if (!SameMemoryStatus( &overall_in_mstat, &overall_out_mstat ))
//		DumpMemoryStatus( "OVERALL MEMORY USAGE", &overall_in_mstat, &overall_out_mstat );
//    }

// simple patch
System::AnsiString	indent = "";
int indenture;
for (indenture=0;indenture<log_indent;indenture++) indent = indent + "    ";

switch (msgclass)
	{
    case	LOG_IN:
    		if (log_indent == 0) new_strings->Add(indent + "-----------------------------------------------------------------------------");
            else new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "IN> " + functiondesc);
			InMemoryStatus( &log_in_mstat );
            log_indent ++;
            break;

    case	LOG_OUT:
			OutMemoryStatus( &log_out_mstat );
            if (!SameMemoryStatus(&log_in_mstat,&log_out_mstat)) DumpMemoryStatus("Function IN/OUT MEMORY STATUS",&log_in_mstat,&log_out_mstat);
			log_indent --;
			indent = "";
			for (indenture=0;indenture<log_indent;indenture++) indent = indent + "    ";
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "OUT> " + functiondesc);
    		if (log_indent != 0) new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
            break;

    case	LOG_INFO:
            if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "INFO> " + functiondesc);
            break;

    case	LOG_NOTE:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc);
            break;

    case	LOG_WARNING:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "WARNING> " + functiondesc + " --> " + msg);
    		else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "WARNING> " + functiondesc);
            break;

    case	LOG_ERROR:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc + " --> " + msg);
    		else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "ERROR> " + functiondesc);
            break;

    case	LOG_MEMSTATUS:
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc + " --> " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "NOTE> " + functiondesc);
            break;

    case	LOG_BEFORE:
    		if (log_indent == 0) new_strings->Add(indent + "-----------------------------------------------------------------------------");
            else new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "BEFORE> " + functiondesc + " --> just before " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "BEFORE> " + functiondesc);
			InMemoryStatus( &log_before_mstat );
            log_indent ++;
            break;

    case	LOG_AFTER:
			log_indent --;
			indent = "";
			for (indenture=0;indenture<log_indent;indenture++) indent = indent + "    ";
    		if (msg != "") new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "AFTER> " + functiondesc + " --> just after " + msg);
            else new_strings->Add(indent + "#" + IntToStr(fErrorCount++) + ":: " + "AFTER> " + functiondesc);
    		if (log_indent != 0) new_strings->Add(indent + " - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ");
			OutMemoryStatus( &log_after_mstat );
            if (!SameMemoryStatus(&log_before_mstat,&log_after_mstat)) DumpMemoryStatus("BEFORE & AFTER " + functiondesc + "::" + msg + " MEMORY STATUS",&log_before_mstat,&log_after_mstat);
            break;
    }
for (int i=0;i<new_strings->Count;i++)
	{
    fErrorStrings->Add(new_strings->Strings[i]);
// why does this not work??
/*
	char* ptr = new_strings->Strings[i].c_str();
    int len = strlen(ptr);
    if (len > 0)
    	{
	    int DebugHandle = open("c:\\temp\\debug.log",O_RDWR|O_APPEND|O_TEXT,S_IREAD|S_IWRITE);
		if (DebugHandle != -1)
	    	{
	        char ret = 0x0D;
	        char lf = 0x0A;
			write(DebugHandle,ptr,len);
	        write(DebugHandle,&ret,1);
	        write(DebugHandle,&lf,1);
	    	close(DebugHandle);
	        }
        }
*/
    }
new_strings->Clear();
delete new_strings;
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDraw::Error( int errorvalue, System::AnsiString functiondesc, System::AnsiString errormsg )
{
Log(LOG_ERROR,functiondesc,"Err # " + IntToStr(errorvalue) + " " + ErrorMessage(errorvalue));
Log(LOG_ERROR,functiondesc,"Err : " + errormsg);
}
//---------------------------------------------------------------------------
System::AnsiString	__fastcall	TDirectDraw::ErrorMessage(int errorvalue)
{
System::AnsiString	errormessage = "";
switch (errorvalue)
	{
    // common directx results
    case DD_OK:						errormessage = "DD_OK, DC retrieved successfully"; break;
	case DDERR_GENERIC:             errormessage = "DDERR_GENERIC, generic error"; break;
	case DDERR_INVALIDOBJECT:		errormessage = "DDERR_INVALIDOBJECT, Invalid object"; break;
	case DDERR_INVALIDPARAMS:		errormessage = "DDERR_INVALIDPARAMS, Invalid params"; break;
	case DDERR_INVALIDSURFACETYPE:	errormessage = "DDERR_INVALIDSURFACETYPE, Invalid surface type"; break;
	case DDERR_UNSUPPORTED:			errormessage = "DDERR_UNSUPPORTED, Unsupported"; break;
	case DDERR_SURFACELOST:			errormessage = "DDERR_SURFACELOST, Surface is lost"; break;
	case DDERR_SURFACEBUSY:			errormessage = "DDERR_SURFACEBUSY, Surface is busy"; break;
	case DDERR_WASSTILLDRAWING:		errormessage = "DDERR_WASSTILLDRAWING, Was still drawing"; break;
	// flip results
	case DDERR_NOFLIPHW:			errormessage = "DDERR_NOFLIPHW, No flip hardware"; break;
	case DDERR_NOTFLIPPABLE:		errormessage = "DDERR_NOTFLIPPABLE, Surface is not flippable"; break;
    // GetDC results
	case DDERR_DCALREADYCREATED:	errormessage = "DDERR_DCALREADYCREATED, DC was already created"; break;
    default:						errormessage = "UNKNOWN ERROR, the error generated is unknown";
    };
return errormessage;
}
//---------------------------------------------------------------------------
/*void	__fastcall	TDirectDraw::Clear(TColor color)
{
Log(LOG_IN,"TDirectDraw::Clear()","color="+IntToStr(color));
if (fBackBufferUsedLast)
	{
	Log(LOG_INFO,"TDirectDraw::Clear()","about to clear the back buffers memory");
	memset( fBackBufferLockingDesc.lpSurface, color, fWidth * fHeight );
	Log(LOG_INFO,"TDirectDraw::Clear()","about to flip the back buffer");
	Flip();
    }
else
    if (fDrawingOnPrimarySurface)
		{
		Log(LOG_INFO,"TDirectDraw::Clear()","about to clear the primary surfaces memory");
		memset( fPrimarySurfaceLockingDesc.lpSurface, color, fWidth * fHeight );
		Log(LOG_INFO,"TDirectDraw::Clear()","about to flip the primary surface");
        Flip();
	    }
Log(LOG_OUT,"TDirectDraw::Clear()","");
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::LockSurface(LPDIRECTDRAWSURFACE surface)
{
if (fDirectDraw == NULL) return;
if (!fEnabled) return false;
Log(LOG_IN,"TDirectDraw::LockSurface()","surface=" + IntToStr((int)surface));

// setup the surface description we use for locking / unlocking surfaces
Log(LOG_NOTE,"TDirectDraw::Constructor()","setting up the locking / unlocking surface descriptions");
memset(&fPrimarySurfaceLockingDesc,0,sizeof(fPrimarySurfaceLockingDesc));
fPrimarySurfaceLockingDesc.dwSize = sizeof(fPrimarySurfaceLockingDesc);
memset(&fBackBufferLockingDesc,0,sizeof(fBackBufferLockingDesc));
fBackBufferLockingDesc.dwSize = sizeof(fBackBufferLockingDesc);

if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Log(LOG_WARNING,"TDirectDraw::LockSurface()","surface cannot be locked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (!fBackBufferLocked)
	    	{
	        Log(LOG_INFO,"TDirectDraw::LockSurface()","about to lock the back buffer");
	        fErrorValue = surface->Lock(NULL,&fBackBufferLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the back buffer could not be locked");
		        return false;
	            }
            fBackBufferLocked = true;
			fBackBufferUsedLast = true;
            fPrimarySurfaceUsedLast = false;
		    Log(LOG_NOTE,"TDirectDraw::LockSurface()","the back buffer locked successfully");
        	}
        else
		    Log(LOG_WARNING,"TDirectDraw::LockSurface()","the back buffer is already locked");
	    }
    else
    	if (!fPrimarySurfaceLocked)
	    	{
    	    Log(LOG_INFO,"TDirectDraw::LockSurface()","about to lock the primary surface");
	        fErrorValue = surface->Lock(NULL,&fPrimarySurfaceLockingDesc,DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,NULL);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::LockSurface()","the primary surface could not be locked");
		        return false;
	            }
            fPrimarySurfaceLocked = true;
            fBackBufferUsedLast = false;
			fPrimarySurfaceUsedLast = true;
		    Log(LOG_NOTE,"TDirectDraw::LockSurface()","the primary surface locked successfully");
	        }
        else
		    Log(LOG_WARNING,"TDirectDraw::LockSurface()","the primary surface is already locked");
	return true;
    }
Log(LOG_OUT,"TDirectDraw::LockSurface()","");
return false;
}
//---------------------------------------------------------------------------
bool	__fastcall	TDirectDraw::UnLockSurface(LPDIRECTDRAWSURFACE surface)
{
if (fDirectDraw == NULL) return;
if (!fEnabled) return false;
Log(LOG_IN,"TDirectDraw::UnLockSurface()","surface=" + IntToStr((int)surface));
if (surface != NULL)
	{
	// note - this function will need to be modified to deal correctly with any surface
	if ((surface != fPrimarySurface) && (surface != fBackBuffer))
    	{
        Log(LOG_WARNING,"TDirectDraw::UnLockSurface()","surface cannot be unlocked because it is neither the primary surface or the back buffer");
        return false;
        }

    if (surface == fBackBuffer)
    	{
        if (fBackBufferLocked)
        	{
	        Log(LOG_INFO,"TDirectDraw::UnLockSurface()","about to lock the back buffer");
			surface->Unlock(fBackBufferLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the back buffer could not be unlocked");
		        return false;
	            }
            fBackBufferLocked = false;
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the back buffer unlocked successfully");
            }
        else
		    Log(LOG_WARNING,"TDirectDraw::UnLockSurface()","the back buffer was already unlocked");
	    }
    else
    	if (fPrimarySurfaceLocked)
	    	{
	        Log(LOG_INFO,"TDirectDraw::UnLockSurface()","about to lock the primary surface");
			fErrorValue = surface->Unlock(fPrimarySurfaceLockingDesc.lpSurface);
			if (fErrorValue != DD_OK)
	        	{
		        Error(fErrorValue,"TDirectDraw::UnLockSurface()","the primary surface could not be unlocked");
		        return false;
	            }
	        fPrimarySurfaceLocked = false;
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface unlocked successfully");
	        }
        else
		    Log(LOG_NOTE,"TDirectDraw::UnLockSurface()","the primary surface was already unlocked");
	return true;
    }
Log(LOG_OUT,"TDirectDraw::UnLockSurface()","");
return false;
}
*/


