//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TDirectDrawForm.h"
//---------------------------------------------------------------------------
#pragma link "TDirectDraw"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TDirectDrawForm *DirectDrawForm;
//---------------------------------------------------------------------------
__fastcall TDirectDrawForm::TDirectDrawForm(TComponent* Owner)
	: TForm(Owner)
{
fAnimationTicks = 0;
fFramesFlipped = 0;
fSeconds = 0;
fColor = 0;

fWndProcDC = NULL;
fHandlingWMPAINT = false;
fHandlingWMDRAWITEM = false;
fControlBeingPainted = -1;

fPaintWithVCL = true;
fPaintWithGDI = false;
fPaintWithDirectDrawDC = false;

fDrawingOnDirectXDC = false;
fDrawingOnDirectXDCWorked = false;
fSaveCanvasHandle = false;
fSaveCanvasHandleWorked = false;
fRestoreCanvasHandle = false;
fRestoreCanvasHandleWorked = false;
fDrawingVCLOnDirectX = false;
fDrawingVCLOnDirectXWorked = false;
fSurfaceWasFlipped = false;

fDirectDraw = new TDirectDraw(this);
if (fDirectDraw != NULL)
	{
    DirectDraw->Debugging = false;
	}
}
//---------------------------------------------------------------------------
__fastcall 	TDirectDrawForm::~TDirectDrawForm()
{
if (fDirectDraw != NULL) delete fDirectDraw;
}
//---------------------------------------------------------------------------
void 	__fastcall TDirectDrawForm::WndProc(Messages::TMessage &Message)
{
// handle any repaint messages coming from Windows rather than letting the default
// WndProc function handle them as the default WndProc function automatically draws
// the form when receiving WM_PAINT, WM_NCPAINT, WM_NCACTIVATE.....
// since the form redrawing code uses a windows gdi Paint3d function which blits to
// the display on completion (we have no control over this), it is probably better
// to use our own handler so that we can control exactly when to blit what.

bool	directx_paint_worked = false;
int		message = Message.Msg;

switch (message)
	{
    case WM_PAINT:
    case WM_NCPAINT:

		// draw background / basic form image
    	if ((fDirectDraw != NULL) && !fHandlingWMPAINT && !fHandlingWMDRAWITEM)
            {
            fHandlingWMPAINT = true;

            if (message == WM_PAINT) DirectDraw->Log(LOG_IN,"<<FORM>>::WndProc()","Received WM_PAINT");
            else DirectDraw->Log(LOG_IN,"<<FORM>>::WndProc()","Received WM_NCPAINT");

	        // this returns the backbuffer if possible, primary surface otherwise
            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->GetDC()");
	      	fWndProcDC = DirectDraw->GetDC();
            DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->GetDC()");

	        if (fWndProcDC != NULL)
	           	{
	            // tell windows that the entire region of the form will need to be repainted
	            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","InvalidateRect()");
	            InvalidateRect(Handle,NULL,false);
    	        DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","InvalidateRect()");

                // perform normal draw functions (without painting the actual form)
                DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","PaintHandler()");
                Messages::TWMPaint	MyPaintMessage;
                MyPaintMessage.Msg = WM_PAINT;
                MyPaintMessage.DC  = fWndProcDC;
                MyPaintMessage.Unused = 0;
                MyPaintMessage.Result = 0;
                PaintHandler(MyPaintMessage);
                DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","PaintHandler()");

			    // make sure to release the DirectX supplied DC
				DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->ReleaseDC()");
				DirectDraw->ReleaseDC(fWndProcDC);
                fWndProcDC = NULL;
				DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->ReleaseDC()");

			    // flip the DirectX buffers (makes recent drawing visible to the user).
				DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->Flip()");
				fSurfaceWasFlipped = DirectDraw->Flip();
				DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->Flip()");
				fFramesFlipped++;

                // set a flag to signal that directx drawing has worked successfully
				DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","!! Direct Drawing Worked !!");
                directx_paint_worked = true;

	            // tell windows that the entire region of the form has been repainted
	            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","ValidateRect()");
	            ValidateRect(fWndProcDC,NULL);
    	        DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","ValidateRect()");
	            }

            DirectDraw->Log(LOG_OUT,"<<FORM>>::WndProc()","");
	        fHandlingWMPAINT = false;
            }

        // make sure that the VCL has finished with the message
        Message.Msg = 0;

        break;

    // note: these messages are sent by windows during the processing of the above messages,
    //   (WM_PAINT and WM_NCPAINT). The WM_PAINT generated during PaintControls() results in
    //   windows being notified & updated immediately, with the WM_PAINT message being
    //   translated to either WM_CTLCOLORBTN or WM_DRAWITEM, depending on the control.
    // response: simply use the same DC as above, as it has been declared static & should contain
    //   the currently accessed DirectX DC

    case WM_NCACTIVATE:

    	// don't do anything as the multiple redraws look ugly...
        Message.Msg = 0;
    	break;

    default:

    	if (fDirectDraw != NULL) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","unknown message " + IntToStr(message));
        break;
    }

// if the directx stuff didn't work correctly, use standard VCL functions
if (!directx_paint_worked)
	{
    TForm::WndProc(Message);
    }
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWindow(HDC hdc)
{
if (hdc == NULL) return;
if (fDirectDraw == NULL)
	{
    TForm::PaintWindow(hdc);
    return;
    }

DirectDraw->Log(LOG_IN,"<<FORM>>::PaintWindow()","");

// save original settings (because we probably want to change them somewhere here)
HDC				orig_canvas_handle = Canvas->Handle;
TColor 			orig_pen_color = Canvas->Pen->Color;
int 			orig_pen_width = Canvas->Pen->Width;
TPenStyle		orig_pen_style = Canvas->Pen->Style;
TColor			orig_brush_color = Canvas->Brush->Color;
TBrushStyle		orig_brush_style = Canvas->Brush->Style;
int 			orig_font_size = Canvas->Font->Size;
TFontName		orig_font_name = Canvas->Font->Name;
TColor			orig_font_color = Canvas->Font->Color;

try
	{
    Canvas->Handle = hdc;
    Paint();
    }
catch ( ... )
	{
    }
// always release anything that might have been allocated in Paint()
Canvas->Handle = NULL;

// restore original settings
Canvas->Handle = orig_canvas_handle;
Canvas->Font->Color = orig_font_color;
Canvas->Font->Name = orig_font_name;
Canvas->Font->Size = orig_font_size;
Canvas->Brush->Style = orig_brush_style;
Canvas->Brush->Color = orig_brush_color;
Canvas->Pen->Style = orig_pen_style;
Canvas->Pen->Width = orig_pen_width;
Canvas->Pen->Color = orig_pen_color;

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintWindow()","");
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintControls(HDC hdc, TControl* First)
{
// simply check our preconditions & skip out if not appropriate
if (fDirectDraw == NULL) return;

// simply check our preconditions & skip out if not appropriate
DirectDraw->Log(LOG_IN,"<<FORM>>::PaintControls()","");
if (ControlCount == 0)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","ControlCount is 0");
    return;
    }
if (hdc == NULL)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","HDC is null");
    return;
	}

// quickly scan the controls until we get to the first control, then fall through with the
// value of the first control in "i" (defaulting at zero if First parameter is NULL)
int i = 0;
int j = 0;
while ((First != NULL) && (j<ControlCount))
   	{
	if (Controls[j] == First) i = j;
    j++;
    }

// create a rectangular region used for resetting the windows clip region (after each
// control is repainted, the clip region is set to that control's clip region - thus we
// reset it to the form's clip region)

HRGN whole_form_clip_rect = CreateRectRgn(Left,Top,Width,Height);
GetClipRgn(hdc,&whole_form_clip_rect);

// set the clip region to that of the form....
SelectClipRgn(hdc,whole_form_clip_rect);
// set the window origin to 0,0 (caJUSTse)
//SetWindowOrg( hdc, 0, 0 );


// scan through all controls, starting with the first of the ordered controls and get them
// to paint to the DirectX supplied DC

int k = 0;
while (k<ControlCount)
	{
    // set a global flag used by the WndProc() function
	fControlBeingPainted = i;

    // use temporary variable
    TControl* control = Controls[i];

    // if we _really_ need to redraw the control
   	if (control->Visible && !control->ControlStyle.Contains(csNoDesignVisible))
    	{
        // debug
		DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Painting Control - '" + control->Name + "'");

		// figure out where we should be drawing the relevant control
		int borderx = GetSystemMetrics(SM_CXDLGFRAME);
		int bordery = GetSystemMetrics(SM_CYDLGFRAME);
        int captionheight = GetSystemMetrics(SM_CYCAPTION);

        int control_dx = 0;
        int control_dy = 0;

        // when required, draw a simple frame around the control....

        if (control->ControlStyle.Contains(csFramed) || !control->ControlStyle.Contains(csOpaque))
        	{
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - has a border, painting border now");

            // update the control offsets (to include border for some bizarre Borland reason)
            control_dx += 2;
            control_dy += 2;

            // paint the appropriate frame
            RECT r1;
	        HBRUSH shadow = CreateSolidBrush(ColorToRGB(clBtnShadow));
	        HBRUSH highlight = CreateSolidBrush(ColorToRGB(clBtnHighlight));

            r1.left = control->Left - 1;
            r1.top = control->Top - 1;
            r1.right = control->Left + control->Width;
            r1.bottom = control->Top + control->Height;
	        FrameRect(hdc, &r1, shadow);

            r1.left++;
            r1.top++;
            r1.right++;
            r1.bottom++;
	        FrameRect(hdc, &r1, highlight);

			DeleteObject(highlight);
            DeleteObject(shadow);
	        }

        // set the canvas clip rect to that of the control
		RECT r2;
		r2.left = borderx + Left + control->Left;
		r2.top  = bordery + captionheight + Top + control->Top;
		r2.right  = r2.left + control->Width;
		r2.bottom = r2.top + control->Height;
		HRGN inside_border_clip_rect = CreateRectRgn( r2.left, r2.top, r2.right, r2.bottom );
		SelectClipRgn( hdc, inside_border_clip_rect );

        // set the window origin to the top,left of the control to be drawn....
        MoveWindowOrg( hdc, control->Left + control_dx, control->Top + control_dy );

		// save current control state flags
        bool paintcopy = (control->ControlState.Contains(csPaintCopy));

        // now make sure the control can be copied
        if (!paintcopy) ControlState << csPaintCopy;

        // check whether or not the control is transparent
        bool transparent = !(control->ControlStyle.Contains(csOpaque));

RECT nr1;
nr1.left = 0;
nr1.top = 0;
nr1.right = control->Width;
nr1.bottom = control->Height;
HBRUSH noticable2 = CreateSolidBrush(ColorToRGB(clYellow));
FrameRect(hdc, &nr1, noticable2);
DeleteObject( noticable2 );

        // only clear the background for controls that are transparent... little bit of
        // a speed increase hopefully
        if (transparent) control->Perform(WM_ERASEBKGND, (long) hdc, NULL);

        // tell the control to repaint itself onto the DirectX supplied DC
       	control->Perform(WM_PAINT, (long) hdc, NULL);

        // log a quick note
		DirectDraw->Log(LOG_NOTE,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - was painted");

        // restore the original state settings
        if (!paintcopy) ControlState >> csPaintCopy;

        // restore the original window origin
        MoveWindowOrg( hdc, -(control->Left + control_dx), -(control->Top + control_dy) );

        // release resources associated with our temporary clipping region
        DeleteObject( inside_border_clip_rect );

		// reset the clip region to what is was before painting this control
		SelectClipRgn( hdc, whole_form_clip_rect );
        }

	// choose the next control
    ++i %= ControlCount;	// wrap to 0..ControlCount-1
    k++;
    }

// reset the global flag used in the WndProc() function.
fControlBeingPainted = -1;

// make sure to blow away our whole_form_clip_rect
DeleteObject( whole_form_clip_rect );

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","painted " + IntToStr(k) + " controls");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintHandler(Messages::TWMPaint &Message)
{
// flag that the draw has not yet worked
fDrawingOnDirectXDCWorked = false;

bool directdraw_created = (fDirectDraw != NULL);
if (directdraw_created && DirectDraw->Enabled)
   	{
	DirectDraw->Log(LOG_IN,"<<FORM>>::PaintHandler()","");

    // paint using standard VCL functions directly on the DirectX supplied surface
	fDrawingVCLOnDirectXWorked = false;

    // if the supplied DC is okay
    HDC hdc = Message.DC;
    if (hdc != NULL)
        // use standard function calls to paint to our DirectX surface
	    try
	    	{
            // paint the form (custom)
            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintForm()");
      		PaintForm(hdc);
            DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintForm()");

		    DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintWindow()");
	        PaintWindow(hdc);
		    DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintWindow()");

			DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintControls()");
		    PaintControls(hdc,NULL);
			DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintControls()");

			fDrawingOnDirectXDCWorked = true;
			DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - SUCCESSFUL!");
	        }
        // unless a problem occurs
		catch (...)
	       	{
	        fDrawingOnDirectXDCWorked = false;
	        DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - NOT SUCCESSFUL!");
	        }
    }

// if the direct drawing did not work, call the standard PaintHandler()
if (!fDrawingOnDirectXDCWorked)
	{
    Message.DC = 0;
    TForm::PaintHandler(Message);
    }
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::Repaint()
{
if (Visible)
	{
    InvalidateRect(Handle,NULL,false);
    UpdateWindow(Handle);
    }
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::PaintForm(HDC hdc)
{
// skip out if invalid DC
if (hdc == NULL) return;
if (fDirectDraw == NULL) return;

DirectDraw->Log(LOG_IN,"<<FORM>>::PaintForm()","");

// save original settings (because we probably want to change them somewhere here)
HDC				orig_canvas_handle = Canvas->Handle;
TColor 			orig_pen_color = Canvas->Pen->Color;
int 			orig_pen_width = Canvas->Pen->Width;
TPenStyle		orig_pen_style = Canvas->Pen->Style;
TColor			orig_brush_color = Canvas->Brush->Color;
TBrushStyle		orig_brush_style = Canvas->Brush->Style;
int 			orig_font_size = Canvas->Font->Size;
TFontName		orig_font_name = Canvas->Font->Name;
TColor			orig_font_color = Canvas->Font->Color;

// set the canvas to paint on the DC
Canvas->Handle = hdc;

// - debugging - fill the entire form area with black
TRect	dr = Canvas->ClipRect;
Canvas->Brush->Color = clBlack;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( dr );

/* various parameters for Win32::GetSystemMetrics()
SM_CXSCREEN           =  0
SM_CYSCREEN           =  1
SM_CXVSCROLL          =  2
SM_CYHSCROLL          =  3
SM_CYCAPTION          =  4
SM_CXBORDER           =  5
SM_CYBORDER           =  6
SM_CXDLGFRAME         =  7
SM_CYDLGFRAME         =  8
SM_CYVTHUMB           =  9
SM_CXHTHUMB           =  10
SM_CXICON             =  11
SM_CYICON             =  12
SM_CXCURSOR           =  13
SM_CYCURSOR           =  14
SM_CYMENU             =  15
SM_CXFULLSCREEN       =  16
SM_CYFULLSCREEN       =  17
SM_CYKANJIWINDOW      =  18
SM_MOUSEPRESENT       =  19
SM_CYVSCROLL          =  20
SM_CXHSCROLL          =  21
SM_DEBUG              =  22
SM_SWAPBUTTON         =  23
SM_RESERVED1          =  24
SM_RESERVED2          =  25
SM_RESERVED3          =  26
SM_RESERVED4          =  27
SM_CXMIN              =  28
SM_CYMIN              =  29
SM_CXSIZE             =  30
SM_CYSIZE             =  31
SM_CXFRAME            =  32
SM_CYFRAME            =  33
SM_CXMINTRACK         =  34
SM_CYMINTRACK         =  35
IFNDEF NOWIN31
SM_CXDOUBLECLK        =  36
SM_CYDOUBLECLK        =  37
SM_CXICONSPACING      =  38
SM_CYICONSPACING      =  39
SM_MENUDROPALIGNMENT  =  40
SM_PENWINDOWS         =  41
SM_DBCSENABLED        =  42
ENDIF
SM_CMETRICSMAX        =  43

*/

// get some stats from windows
int borderx = GetSystemMetrics(SM_CXDLGFRAME);
int bordery = GetSystemMetrics(SM_CYDLGFRAME);
int captionheight = GetSystemMetrics(SM_CYCAPTION);
int x,y;

// draw a 3d frame round the entire form
Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;

// left edge
for (x=0;x<borderx;x++)
	{
	if (x==0) Canvas->Pen->Color = clBtnHighlight;
    else Canvas->Pen->Color = clBtnFace;
	Canvas->MoveTo(x,0+x);
	Canvas->LineTo(x,Height-x+bordery);
    }
// top edge
for (y=0;y<bordery;y++)
    {
	if (y==0) Canvas->Pen->Color = clBtnHighlight;
    else Canvas->Pen->Color = clBtnFace;
    Canvas->MoveTo(0+y,y);
    Canvas->LineTo(Width-y,y);
    }
// right edge
for (x=0;x<borderx;x++)
    {
	if (x==0) Canvas->Pen->Color = clBtnShadow;
    else Canvas->Pen->Color = clBtnFace;
	Canvas->MoveTo(Width-x-1,0+x);
	Canvas->LineTo(Width-x-1,Height-x+bordery);
    }
// bottom edge
for (y=0;y<bordery;y++)
    {
	if (y==0) Canvas->Pen->Color = clBtnShadow;
    else Canvas->Pen->Color = clBtnFace;
    Canvas->MoveTo(0+y,Height-y+bordery);
    Canvas->LineTo(Width-y,Height-y+bordery);
    }

// fill in the caption area with the active caption color
TRect	r = Canvas->ClipRect;
TRect	caption_r = r;
caption_r.Left += borderx;
caption_r.Top  += bordery;
caption_r.Right -= borderx;
caption_r.Bottom = caption_r.Top + captionheight + 2;
Canvas->Brush->Color = clActiveCaption;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( caption_r );

// draw line between caption bar and main draw area
Canvas->Pen->Color = clBtnFace;
Canvas->Pen->Width = bordery;
Canvas->MoveTo(caption_r.Left, caption_r.Bottom + 1);
Canvas->LineTo(caption_r.Right, caption_r.Bottom + 1);

// draw any buttons on the caption bar



// draw the form's caption
Canvas->Brush->Color = clActiveCaption;
Canvas->Font->Size = Font->Size;
Canvas->Font->Color = Font->Color;
Canvas->Font->Name = Font->Name;
Canvas->TextOut( caption_r.Left+1,caption_r.Top+1, Caption );
if (fDirectDraw != NULL)
	{
    if (DirectDraw->Enabled)
    	{
		Canvas->Font->Color = clGreen;
		Canvas->TextOut( caption_r.Left+1+600,caption_r.Top+1, "DirectX is Enabled" );
        }
    else
    	{
		Canvas->Font->Color = clRed;
		Canvas->TextOut( caption_r.Left+1+600,caption_r.Top+1, "DirectX is Not Enabled" );
        }
    }
else
	{
	Canvas->Font->Color = clRed;
	Canvas->TextOut( caption_r.Left+1+600,caption_r.Top+1, "DirectX is Not Available" );
    }

// fill the main form area with it's color
TRect	main_r;
main_r.Left = caption_r.Left;
main_r.Top  = caption_r.Bottom + bordery;
main_r.Right = caption_r.Right;
main_r.Bottom = r.Bottom - bordery;
Canvas->Brush->Color = Color;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( main_r );

// always release anything that may have been allocated by the Canvas
Canvas->Handle = NULL;

// restore original settings
Canvas->Handle = orig_canvas_handle;
Canvas->Font->Color = orig_font_color;
Canvas->Font->Name = orig_font_name;
Canvas->Font->Size = orig_font_size;
Canvas->Brush->Style = orig_brush_style;
Canvas->Brush->Color = orig_brush_color;
Canvas->Pen->Style = orig_pen_style;
Canvas->Pen->Width = orig_pen_width;
Canvas->Pen->Color = orig_pen_color;

// make sure that the window origin is inside the borders / caption area
MoveWindowOrg(hdc, main_r.Left, main_r.Top);

// make sure that the clip rectangle is inside the borders / caption area (0,0 since origin has moved)
HRGN inside_border_clip_rect = CreateRectRgn(0,0,main_r.Right-main_r.Left,main_r.Bottom-main_r.Top);
SelectClipRgn( Canvas->Handle, inside_border_clip_rect );
DeleteObject( inside_border_clip_rect );

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintForm()","");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ExitButtonClick(TObject *Sender)
{
if (Enabled) Close();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintWithVCLButtonClick(TObject *Sender)
{
fPaintWithVCL = !fPaintWithVCL;
fPaintWithGDI = false;
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintWithGDIButtonClick(TObject *Sender)
{
fPaintWithGDI = !fPaintWithGDI;
fPaintWithVCL = false;
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintWithDirectDrawButtonClick(TObject *Sender)
{
fPaintWithDirectDrawDC = !fPaintWithDirectDrawDC;
DirectDraw->Enabled = fPaintWithDirectDrawDC;
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::UpdateDirectXMessages()
{
if (DirectDraw->ErrorStrings)
	{
	DirectDrawErrorMessagesBox->Items->Clear();
	for (int i=0;i<DirectDraw->ErrorStrings->Count;i++)
		DirectDrawErrorMessagesBox->Items->Add(DirectDraw->ErrorStrings->Strings[i]);
	DirectDraw->ErrorCount = 0;
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormCreate(TObject *Sender)
{
bool all_enabled = true;
DirectDraw->Enabled = true;
if (DirectDraw->Enabled)	
	{
	if (DirectDraw->SetDisplayMode(640,480,8))
    	{
	    AnimationTimer->Interval = AutoFlipTimer->Interval;
	    AnimationTimer->Enabled = true;
	    AutoFlipTimer->Enabled = true;
        all_enabled = true;
        }
    }
else
	ShowMessage("DirectDraw could not be enabled");

if (!all_enabled)
	{
    for (int i=0;i<ControlCount;i++) Controls[i]->Enabled = false;
    AnimationTimer->Enabled = false;
    AutoFlipTimer->Enabled = false;
    }

Left=0; Width=DirectDraw->Width-Left;
Top=5; Height=DirectDraw->Height-Top;
DirectDrawErrorMessagesBox->Height -= 5;

fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormActivate(TObject *Sender)
{
AnimationTimer->Enabled = true;
if (DirectDraw->ErrorValue == DD_OK) DirectXInstalledLabel->Caption = "DirectX is Installed";
else DirectXInstalledLabel->Caption = "DirectX is Not Installed";
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormDeactivate(TObject *Sender)
{
AnimationTimer->Enabled = false;
AutoFlipTimer->Enabled = false;
DirectDraw->Enabled = false;
DirectDrawErrorMessagesBox->Items->Clear();
}
//---------------------------------------------------------------------------
TDateTime last_time = Time();
int PageNumber;
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormPaint(TObject *Sender)
{
TDateTime current_time = Time();
unsigned short chour,cmin,csec,cmsec;
unsigned short lhour,lmin,lsec,lmsec;
current_time.DecodeTime(&chour,&cmin,&csec,&cmsec);
last_time.DecodeTime(&lhour,&lmin,&lsec,&lmsec);
int seconds_elapsed = 0;
if (csec != lsec)
	{
    if (csec < lsec) seconds_elapsed += 60;
    seconds_elapsed += (csec - lsec);
    }
fSeconds += seconds_elapsed;
last_time = current_time;

++PageNumber %= 2;	// 0..1 only

Canvas->Pen->Color = clRed;
Canvas->Font->Color = clWhite;
Canvas->TextOut(200,0,"PAGE # " + IntToStr(PageNumber));

// BEGIN REGULAR DRAW CODE
TicksLabel->Caption = "Ticks: " + IntToStr(fAnimationTicks);
FramesLabel->Caption = "Frames: " + IntToStr(fFramesFlipped);
SecondsLabel->Caption = "Seconds: " + IntToStr(fSeconds);
if (fSeconds != 0)
	{
    FramesPerSecondLabel->Caption = "FPS: " + FloatToStrF((float)fFramesFlipped / (float)fSeconds, ffFixed, 6, 2);
	TicksPerSecondLabel->Caption = "TPS: " + FloatToStrF((float)fAnimationTicks / (float)fSeconds, ffFixed, 6, 2);
	}

if (fPaintWithVCL)
	{
	if (fPaintWithDirectDrawDC)
    	{
		PaintWithVCLButton->Caption = "Painting with &VCL on DIRECTX surface: ON";
        PaintWithLabel->Caption = "Using standard VCL functions to paint on a DirectX surface";
		PaintWithLabel->Font->Color = clGreen;
        }
    else
    	{
        PaintWithVCLButton->Caption = "Paint with &VCL on VCL canvas: ON";
        PaintWithLabel->Caption = "Using standard VCL functions to paint on a VCL canvas";
		PaintWithLabel->Font->Color = clRed;
        }
    PaintWithVCL();
    }
else
	{
    if (fPaintWithDirectDrawDC)	PaintWithVCLButton->Caption = "Paint with &VCL on DIRECTX surface: OFF";
	else PaintWithVCLButton->Caption = "Paint with &VCL on VCL canvas: OFF";
    }
if (fPaintWithGDI)
	{
	if (fPaintWithDirectDrawDC)
    	{
        PaintWithGDIButton->Caption = "Paint with &GDI on DIRECTX surface: ON";
        PaintWithLabel->Caption = "Using standard GDI functions to paint on a DirectX surface";
		PaintWithLabel->Font->Color = clGreen;
        }
    else
    	{
        PaintWithGDIButton->Caption = "Paint with &GDI on VCL canvas: ON";
        PaintWithLabel->Caption = "Using standard GDI functions to paint on a VCL canvas";
		PaintWithLabel->Font->Color = clRed;
        }
    PaintWithGDI();
    }
else
	{
	if (fPaintWithDirectDrawDC) PaintWithGDIButton->Caption = "Paint with &GDI on DIRECTX surface: OFF";
    else PaintWithGDIButton->Caption = "Paint with &GDI on VCL canvas: OFF";
    }
if (fPaintWithDirectDrawDC)PaintWithDirectDrawButton->Caption = "Paint directly on &DirectX surface: ON";
else PaintWithDirectDrawButton->Caption = "Paint directly on &DirectX surface: OFF";

if (DirectDraw->Debugging)
	{
    DebuggingLabel->Font->Color = clGreen;
    DebuggingLabel->Caption = "DirectX debugging is ON";
    }
else
	{
    DebuggingLabel->Font->Color = clRed;
    DebuggingLabel->Caption = "DirectX debugging is OFF";
    }

if (fDrawingOnDirectXDC)
	{
    DrawingOnDirectXDCLabel->Font->Color = clGreen;
    DrawingOnDirectXDCLabel->Caption = "drawing on DirectX supplied surface";
    }
else
	{
    DrawingOnDirectXDCLabel->Font->Color = clRed;
    DrawingOnDirectXDCLabel->Caption = "drawing on VCL supplied surface";
    }

if (fDrawingOnDirectXDCWorked)
	{
    DrawingOnDirectXDCWorkedLabel->Font->Color = clGreen;
    DrawingOnDirectXDCWorkedLabel->Caption = "drawing on DirectX surface worked";
    }
else
	{
    DrawingOnDirectXDCWorkedLabel->Font->Color = clRed;
    DrawingOnDirectXDCWorkedLabel->Caption = "drawing on DirectX surface failed";
    }

if (fSaveCanvasHandle)
	{
    SaveCanvasHandleLabel->Font->Color = clGreen;
    SaveCanvasHandleLabel->Caption = "canvas handle was saved (freed)";
    }
else
	{
    SaveCanvasHandleLabel->Font->Color = clRed;
    SaveCanvasHandleLabel->Caption = "canvas handle was not saved (freed)";
    }

if (fSaveCanvasHandleWorked)
	{
    SaveCanvasHandleWorkedLabel->Font->Color = clGreen;
    SaveCanvasHandleWorkedLabel->Caption = "canvas handle was saved (freed) correctly";
    }
else
	{
    SaveCanvasHandleWorkedLabel->Font->Color = clRed;
    SaveCanvasHandleWorkedLabel->Caption = "canvas handle did not save (free) correctly";
    }

if (fRestoreCanvasHandle)
	{
    RestoreCanvasHandleLabel->Font->Color = clGreen;
    RestoreCanvasHandleLabel->Caption = "canvas handle was restored";
    }
else
	{
    RestoreCanvasHandleLabel->Font->Color = clRed;
    RestoreCanvasHandleLabel->Caption = "canvas handle was not not restored";
    }

if (fRestoreCanvasHandleWorked)
	{
    RestoreCanvasHandleWorkedLabel->Font->Color = clGreen;
    RestoreCanvasHandleWorkedLabel->Caption = "canvas handle was restored (zeroed) correctly";
    }
else
	{
    RestoreCanvasHandleWorkedLabel->Font->Color = clRed;
    RestoreCanvasHandleWorkedLabel->Caption = "canvas handle did not restore (zero) correctly";
    }

if (fDrawingVCLOnDirectX)
	{
    DrawingVCLOnDirectXLabel->Caption = "attempting to use VCL to draw on DirectX DC";
    }
else
	{
    DrawingVCLOnDirectXLabel->Caption = "not using VCL to draw on DirectX DC ?? reason?";
    }

if (fDrawingVCLOnDirectXWorked)
	{
    DrawingVCLOnDirectXWorkedLabel->Caption = "Drawing using VCL on DirectX surface worked";
    }
else
	{
    DrawingVCLOnDirectXWorkedLabel->Caption = "Drawing using VCL on DirectX not working";
    }

if (fSurfaceWasFlipped)
	{

    SurfaceWasFlippedLabel->Caption = "surface was flipped";
    }
else
	{
    SurfaceWasFlippedLabel->Caption = "surface was not flipped";
    }
}
//---------------------------------------------------------------------------
int x = 0;
int y = 0;
int xincr = 1;
int yincr = 1;
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWithVCL()
{
// animation code
fColor++;
x += xincr;
y += yincr;

if (x<0) xincr = 1;
if (y<0) yincr = 1;
if (x+400>Width) xincr = -1;
if (y+200>Height) yincr = -1;

Canvas->Font->Color = clWhite;
Canvas->TextOut(700,0,"X == " + IntToStr(x));

// use the standard VCL functions to draw on the DirectX surface as we see fit
Canvas->Pen->Color = (TColor) RGB(255,255,255);
Canvas->Brush->Color = (TColor) RGB((x*128)%256,(x*16)%256,(x*16+16*fColor)%256) ;
Canvas->Brush->Style = bsSolid;
Canvas->Rectangle( x,y, x+400,y+200 );

UpdateDirectXMessages();
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWithGDI()
{
// use standard GDI functions to manipulate the DirectX surface as we see fit
SetTextColor( Canvas->Handle, RGB(255,255,255));
SetBkColor( Canvas->Handle, RGB((x*128)%256,(x*16)%256,(x*16+16*fColor)%256));
RECT r;
r.left = x;
r.top = y;
r.right = x+400;
r.bottom = y+200;
FillRect( Canvas->Handle, &r, 0 );
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn1Click(TObject *Sender)
{
Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn2Click(TObject *Sender)
{
if (Enabled)
	{
    DirectDraw->ErrorStrings->Clear();
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn3Click(TObject *Sender)
{
if (Enabled)
	{
    DirectDraw->Debugging = !DirectDraw->Debugging;
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::SecondsTimerTimer(TObject *Sender)
{
fSeconds++;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::AnimationTimerTimer(TObject *Sender)
{
fAnimationTicks++;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::AutoFlipTimerTimer(TObject *Sender)
{
fFramesFlipped++;
Refresh();
}
//---------------------------------------------------------------------------
// remember: function SetPriorityClass(hProcess: THandle; dwPriorityClass: DWORD): BOOL; stdcall;

void __fastcall TDirectDrawForm::BitBtn4Click(TObject *Sender)
{
if (Enabled) if (AnimationTimer->Interval>20) AnimationTimer->Interval -= 20;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn5Click(TObject *Sender)
{
if (Enabled) if (AnimationTimer->Interval<2000) AnimationTimer->Interval += 20;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormDestroy(TObject *Sender)
{
DirectDrawErrorMessagesBox->Items->Clear();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::SpeedButton1Click(TObject *Sender)
{
for (int i=0;i<DirectDrawErrorMessagesBox->Items->Count;i++)
	{
    if (AnsiPos(DirectDrawErrorMessagesBox->Items->Strings[i],Edit3->Text) != 0)
    	{
    	DirectDrawErrorMessagesBox->ItemIndex = i;
        break;
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::SpeedButton2Click(TObject *Sender)
{
for (int i=DirectDrawErrorMessagesBox->ItemIndex;i<DirectDrawErrorMessagesBox->Items->Count;i++)
	{
    if (AnsiPos(DirectDrawErrorMessagesBox->Items->Strings[i], Edit3->Text) != 0)
    	{
    	DirectDrawErrorMessagesBox->ItemIndex = i;
        break;
        }
    }
}
//---------------------------------------------------------------------------
