//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TDirectDrawForm.h"
//---------------------------------------------------------------------------
#pragma link "TDirectDraw"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TDirectDrawForm *DirectDrawForm;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// remember: function SetPriorityClass(hProcess: THandle; dwPriorityClass: DWORD): BOOL; stdcall;
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
__fastcall TDirectDrawForm::TDirectDrawForm(TComponent* Owner)
	: TForm(Owner)
{
fAnimationTicks = 0;
fFramesFlipped = 0;
fSeconds = 0;
fColor = 0;

fWndProcDC = NULL;
fHandlingWMPAINT = false;
fControlBeingPainted = -1;

fDirectDraw = new TDirectDraw(this);
if (fDirectDraw != NULL)
	{
    DirectDraw->Debugging = false;
	}
}
//---------------------------------------------------------------------------
__fastcall 	TDirectDrawForm::~TDirectDrawForm()
{
if (fDirectDraw != NULL) delete fDirectDraw;
}
//---------------------------------------------------------------------------
void 	__fastcall TDirectDrawForm::WndProc(Messages::TMessage &Message)
{
// handle any repaint messages coming from Windows rather than letting the default
// WndProc function handle them as the default WndProc function automatically draws
// the form when receiving WM_PAINT, WM_NCPAINT, WM_NCACTIVATE.....
// since the form redrawing code uses a windows gdi Paint3d function which blits to
// the display on completion (we have no control over this), it is probably better
// to use our own handler so that we can control exactly when to blit what.

bool	directx_paint_worked = false;
int		message = Message.Msg;

switch (message)
	{
    case WM_PAINT:
    case WM_NCPAINT:

		// draw background / basic form image
    	if ((fDirectDraw != NULL) && !fHandlingWMPAINT)
            {
            fHandlingWMPAINT = true;

            if (message == WM_PAINT) DirectDraw->Log(LOG_IN,"<<FORM>>::WndProc()","Received WM_PAINT");
            else DirectDraw->Log(LOG_IN,"<<FORM>>::WndProc()","Received WM_NCPAINT");

	        // this returns the backbuffer if possible, primary surface otherwise
            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->GetDC()");
	      	fWndProcDC = DirectDraw->GetDC();
            DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->GetDC()");

	        if (fWndProcDC != NULL)
	           	{
	            // tell windows that the entire region of the form will need to be repainted
	            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","InvalidateRect()");
	            InvalidateRect(Handle,NULL,false);
    	        DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","InvalidateRect()");

                // perform normal draw functions (without painting the actual form)
                DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","PaintHandler()");
                Messages::TWMPaint	MyPaintMessage;
                MyPaintMessage.Msg = WM_PAINT;
                MyPaintMessage.DC  = fWndProcDC;
                MyPaintMessage.Unused = 0;
                MyPaintMessage.Result = 0;
                PaintHandler(MyPaintMessage);
                DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","PaintHandler()");

			    // make sure to release the DirectX supplied DC
				DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->ReleaseDC()");
				DirectDraw->ReleaseDC(fWndProcDC);
                fWndProcDC = NULL;
				DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->ReleaseDC()");

			    // flip the DirectX buffers (makes recent drawing visible to the user).
				DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","DirectDraw->Flip()");
				if (DirectDraw->Flip()) fFramesFlipped++;
				DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","DirectDraw->Flip()");

                // set a flag to signal that directx drawing has worked successfully
				DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","!! Direct Drawing Worked !!");
                directx_paint_worked = true;

	            // tell windows that the entire region of the form has been repainted
	            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::WndProc()","ValidateRect()");
	            ValidateRect(fWndProcDC,NULL);
    	        DirectDraw->Log(LOG_AFTER,"<<FORM>>::WndProc()","ValidateRect()");
	            }

            DirectDraw->Log(LOG_OUT,"<<FORM>>::WndProc()","");
	        fHandlingWMPAINT = false;
            }

        // make sure that the VCL has finished with the message
        Message.Msg = 0;

        break;

    case WM_NCACTIVATE:

    	// don't do anything as the multiple redraws look ugly...
        Message.Msg = 0;
    	break;

    default:

    	if (fDirectDraw != NULL) DirectDraw->Log(LOG_NOTE,"<<FORM>>::WndProc()","unknown message " + IntToStr(message));
        break;
    }

// if the directx stuff didn't work correctly, use standard VCL functions
if (!directx_paint_worked)
	{
    TForm::WndProc(Message);
    }
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintWindow(HDC hdc)
{
if (hdc == NULL) return;
if (fDirectDraw == NULL)
	{
    TForm::PaintWindow(hdc);
    return;
    }

DirectDraw->Log(LOG_IN,"<<FORM>>::PaintWindow()","");

// save original settings (because we probably want to change them somewhere here)
HDC				orig_canvas_handle = Canvas->Handle;
TColor 			orig_pen_color = Canvas->Pen->Color;
int 			orig_pen_width = Canvas->Pen->Width;
TPenStyle		orig_pen_style = Canvas->Pen->Style;
TColor			orig_brush_color = Canvas->Brush->Color;
TBrushStyle		orig_brush_style = Canvas->Brush->Style;
int 			orig_font_size = Canvas->Font->Size;
TFontName		orig_font_name = Canvas->Font->Name;
TColor			orig_font_color = Canvas->Font->Color;

try
	{
    Canvas->Handle = hdc;
    Paint();
    }
catch ( ... )
	{
    }
// always release anything that might have been allocated in Paint()
Canvas->Handle = NULL;

// restore original settings
Canvas->Handle = orig_canvas_handle;
Canvas->Font->Color = orig_font_color;
Canvas->Font->Name = orig_font_name;
Canvas->Font->Size = orig_font_size;
Canvas->Brush->Style = orig_brush_style;
Canvas->Brush->Color = orig_brush_color;
Canvas->Pen->Style = orig_pen_style;
Canvas->Pen->Width = orig_pen_width;
Canvas->Pen->Color = orig_pen_color;

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintWindow()","");
}
//---------------------------------------------------------------------------
void	__fastcall	TDirectDrawForm::PaintControls(HDC hdc, TControl* First)
{
// simply check our preconditions & skip out if not appropriate
if (fDirectDraw == NULL) return;

// simply check our preconditions & skip out if not appropriate
DirectDraw->Log(LOG_IN,"<<FORM>>::PaintControls()","");
if (ControlCount == 0)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","ControlCount is 0");
    return;
    }
if (hdc == NULL)
	{
	DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","HDC is null");
    return;
	}

// quickly scan the controls until we get to the first control, then fall through with the
// value of the first control in "i" (defaulting at zero if First parameter is NULL)
// ; also determine how many controls are meant to be repainted
int controls_to_paint = ControlCount;
int i = 0;
int j = 0;
if (First != NULL)
	{
    controls_to_paint = 1;
	while (j<ControlCount)
	   	{
		if (Controls[j] == First) i = j;
	    j++;
	    }
    }

// create a rectangular region used for resetting the windows clip region (after each
// control is repainted, the clip region is set to that control's clip region - thus we
// reset it to the form's clip region)

HRGN whole_form_clip_rect = CreateRectRgn(Left,Top,Width,Height);
GetClipRgn(hdc,&whole_form_clip_rect);

// set the clip region to that of the form....
SelectClipRgn(hdc,whole_form_clip_rect);
// set the window origin to 0,0 (caJUSTse)
//SetWindowOrg( hdc, 0, 0 );


// scan through all controls, starting with the first of the ordered controls and get them
// to paint to the DirectX supplied DC

int k = 0;
while (k<controls_to_paint)
	{
    // set a global flag used by the WndProc() function
	fControlBeingPainted = i;

    // use mapped variables
    TControl* control = Controls[i];

    // figure out if the class is based on TWinControl (if so, requires a border).
    // this is done by traversing the ancestor classes of the control until we find a
    // TWinControl parent, or till we run out of parents
    TClass	theclass = control->ClassType();
    bool isWinControl = false;
	while	(theclass != NULL)
		{
	    System::AnsiString	class_name = theclass->ClassName();
		DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' ancestors - '" + class_name + "'");
        if (class_name == "TWinControl")
        	{
            isWinControl = true;
            break;
            }
		theclass = theclass->ClassParent();
		}

    // if we _really_ need to redraw the control
   	if (control->Visible && !control->ControlStyle.Contains(csNoDesignVisible))
    	{
        // debug
		DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Painting Control - '" + control->Name + "' of class - '" + control->ClassName() + "'");

		// figure out where we should be drawing the relevant control
		int borderx = GetSystemMetrics(SM_CXDLGFRAME);
		int bordery = GetSystemMetrics(SM_CYDLGFRAME);
        int captionheight = GetSystemMetrics(SM_CYCAPTION);

        int control_dx = 0;
        int control_dy = 0;

        // when required, draw a simple frame around the control....

        if (control->ControlStyle.Contains(csFramed) || (!control->ControlStyle.Contains(csOpaque) && isWinControl))
        	{
			DirectDraw->Log(LOG_INFO,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - has a border, painting border now");

            // update the control offsets (to include border for some bizarre Borland reason)
            control_dx += 2;
            control_dy += 2;

            // paint the appropriate frame
            RECT r1;
	        HBRUSH shadow = CreateSolidBrush(ColorToRGB(clBtnShadow));
	        HBRUSH highlight = CreateSolidBrush(ColorToRGB(clBtnHighlight));

            r1.left = control->Left;
            r1.top = control->Top;
            r1.right = control->Left + control->Width;
            r1.bottom = control->Top + control->Height;
	        FrameRect(hdc, &r1, shadow);

            r1.left++;
            r1.top++;
            r1.right++;                              
            r1.bottom++;
	        FrameRect(hdc, &r1, highlight);

			DeleteObject(highlight);
            DeleteObject(shadow);
	        }

        // set the canvas clip rect to that of the control
		RECT r2;
		r2.left = borderx + Left + control->Left;
		r2.top  = bordery + captionheight + Top + control->Top;
		r2.right  = r2.left + control->Width;
		r2.bottom = r2.top + control->Height;
		HRGN inside_border_clip_rect = CreateRectRgn( r2.left, r2.top, r2.right, r2.bottom );
		SelectClipRgn( hdc, inside_border_clip_rect );

        // set the window origin to the top,left of the control to be drawn....
        MoveWindowOrg( hdc, control->Left + control_dx, control->Top + control_dy );

// draw debugging lines
//RECT nr1;
//nr1.left = 0;
//nr1.top = 0;
//nr1.right = control->Width;
//nr1.bottom = control->Height;
//HBRUSH noticable2 = CreateSolidBrush(ColorToRGB(clYellow));
//FrameRect(hdc, &nr1, noticable2);
//DeleteObject( noticable2 );

        // only clear the background for controls that are transparent... little bit of
        // a speed increase hopefully
        if (!control->ControlStyle.Contains(csOpaque)) control->Perform(WM_ERASEBKGND, (long) hdc, NULL);

        // tell the control to repaint itself onto the DirectX supplied DC
       	control->Perform(WM_PAINT, (long) hdc, NULL);

        // log a quick note
		DirectDraw->Log(LOG_NOTE,"<<FORM>>::PaintControls()","Control - '" + control->Name + "' - was painted");

        // restore the original window origin
        MoveWindowOrg( hdc, -(control->Left + control_dx), -(control->Top + control_dy) );

        // release resources associated with our temporary clipping region
        DeleteObject( inside_border_clip_rect );

		// reset the clip region to what is was before painting this control
		SelectClipRgn( hdc, whole_form_clip_rect );
        }

	// choose the next control
    ++i %= ControlCount;	// wrap to 0..ControlCount-1
    k++;
    }

// reset the global flag used in the WndProc() function.
fControlBeingPainted = -1;

// make sure to blow away our whole_form_clip_rect
DeleteObject( whole_form_clip_rect );

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintControls()","painted " + IntToStr(k) + " controls");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::PaintHandler(Messages::TWMPaint &Message)
{
if (fDirectDraw == NULL) return;

bool drawing_worked = false;
if (DirectDraw->Enabled)
   	{
	DirectDraw->Log(LOG_IN,"<<FORM>>::PaintHandler()","");

    // if the supplied DC is okay
    HDC hdc = Message.DC;
    if (hdc != NULL)
        // use standard function calls to paint to our DirectX surface
	    try
	    	{
            // paint the form (custom)
            DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintForm()");
      		PaintForm(hdc);
            DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintForm()");

		    DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintWindow()");
	        PaintWindow(hdc);
		    DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintWindow()");

			DirectDraw->Log(LOG_BEFORE,"<<FORM>>::PaintHandler()","PaintControls()");
		    PaintControls(hdc,NULL);
			DirectDraw->Log(LOG_AFTER,"<<FORM>>::PaintHandler()","PaintControls()");

            drawing_worked = true;
			DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - SUCCESSFUL!");
	        }
        // unless a problem occurs
		catch (...)
	       	{
	        DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintHandler()","VCL paint on DirectX DC - NOT SUCCESSFUL!");
	        }
    }

// if the direct drawing did not work, call the standard PaintHandler() (and make it supply a DC)
if (!drawing_worked)
	{
    Message.DC = 0;
    TForm::PaintHandler(Message);
    }
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::Repaint()
{
if (Visible)
	{
    InvalidateRect(Handle,NULL,false);
    UpdateWindow(Handle);
    }
}
//---------------------------------------------------------------------------
void	__fastcall TDirectDrawForm::PaintForm(HDC hdc)
{
// skip out if invalid DC
if (hdc == NULL) return;
if (fDirectDraw == NULL) return;

DirectDraw->Log(LOG_IN,"<<FORM>>::PaintForm()","");

// save original settings (because we probably want to change them somewhere here)
HDC				orig_canvas_handle = Canvas->Handle;
TColor 			orig_pen_color = Canvas->Pen->Color;
int 			orig_pen_width = Canvas->Pen->Width;
TPenStyle		orig_pen_style = Canvas->Pen->Style;
TColor			orig_brush_color = Canvas->Brush->Color;
TBrushStyle		orig_brush_style = Canvas->Brush->Style;
int 			orig_font_size = Canvas->Font->Size;
TFontName		orig_font_name = Canvas->Font->Name;
TColor			orig_font_color = Canvas->Font->Color;

// set the canvas to paint on the DC
Canvas->Handle = hdc;

// - debugging - fill the entire form area with black
TRect	dr = Canvas->ClipRect;
Canvas->Brush->Color = clYellow;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( dr );

// get some stats from windows
int borderx = GetSystemMetrics(SM_CXDLGFRAME);
int bordery = GetSystemMetrics(SM_CYDLGFRAME);
int captionheight = GetSystemMetrics(SM_CYCAPTION);
int x,y;

// draw a 3d frame round the entire form
Canvas->Pen->Width = 1;
Canvas->Pen->Style = psSolid;

// left & right edges
for (x=0;x<borderx;x++)
	{
	if (x==0) Canvas->Pen->Color = clBtnHighlight;
    else Canvas->Pen->Color = clBtnFace;
	Canvas->MoveTo(x,0+x);
	Canvas->LineTo(x,Height+bordery+1);
	Canvas->MoveTo(Width-x-1,0+x);
	Canvas->LineTo(Width-x-1,Height+bordery+1);
    }
// top & bottom edges
for (y=0;y<bordery;y++)
    {
	if (y==0) Canvas->Pen->Color = clBtnHighlight;
    else Canvas->Pen->Color = clBtnFace;
    Canvas->MoveTo(0+y,y);
    Canvas->LineTo(Width-y,y);
    Canvas->MoveTo(0+y,Height-y+bordery+1);
    Canvas->LineTo(Width-y,Height-y+bordery+1);
    }

// fill in the caption area with the active caption color
TRect	r = Canvas->ClipRect;
TRect	caption_r = r;
caption_r.Left += borderx;
caption_r.Top  += bordery;
caption_r.Right -= borderx;
caption_r.Bottom = caption_r.Top + captionheight + bordery;
Canvas->Brush->Color = clActiveCaption;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( caption_r );

// draw line between caption bar and main draw area
Canvas->Pen->Color = clBtnFace;
Canvas->Pen->Width = bordery;
Canvas->MoveTo(caption_r.Left, caption_r.Bottom + 1);
Canvas->LineTo(caption_r.Right, caption_r.Bottom + 1);

// draw any buttons on the caption bar



// draw the form's caption
Canvas->Brush->Color = clActiveCaption;
Canvas->Font->Size = Font->Size;
Canvas->Font->Color = Font->Color;
Canvas->Font->Name = Font->Name;
Canvas->TextOut( caption_r.Left+borderx,caption_r.Top, Caption );
if (fDirectDraw != NULL)
	{
    if (DirectDraw->Enabled)
    	{
		Canvas->Font->Color = clGreen;
		Canvas->TextOut( caption_r.Left+600,caption_r.Top, "DirectX is Enabled" );
        }
    else
    	{
		Canvas->Font->Color = clRed;
		Canvas->TextOut( caption_r.Left+600,caption_r.Top, "DirectX is Not Enabled" );
        }
    }
else
	{
	Canvas->Font->Color = clRed;
	Canvas->TextOut( caption_r.Left+600,caption_r.Top, "DirectX is Not Available" );
    }

// fill the main form area with it's color
TRect	main_r;
main_r.Left = caption_r.Left;
main_r.Top  = caption_r.Bottom + bordery;
main_r.Right = caption_r.Right;
main_r.Bottom = r.Bottom - bordery;
Canvas->Brush->Color = Color;
Canvas->Brush->Style = bsSolid;
Canvas->FillRect( main_r );

// always release anything that may have been allocated by the Canvas
Canvas->Handle = NULL;

// restore original settings
Canvas->Handle = orig_canvas_handle;
Canvas->Font->Color = orig_font_color;
Canvas->Font->Name = orig_font_name;
Canvas->Font->Size = orig_font_size;
Canvas->Brush->Style = orig_brush_style;
Canvas->Brush->Color = orig_brush_color;
Canvas->Pen->Style = orig_pen_style;
Canvas->Pen->Width = orig_pen_width;
Canvas->Pen->Color = orig_pen_color;

// make sure that the window origin is inside the borders / caption area
MoveWindowOrg(hdc, main_r.Left, main_r.Top);

// make sure that the clip rectangle is inside the borders / caption area (0,0 since origin has moved)
HRGN inside_border_clip_rect = CreateRectRgn(0,0,main_r.Right-main_r.Left,main_r.Bottom-main_r.Top);
SelectClipRgn( Canvas->Handle, inside_border_clip_rect );
DeleteObject( inside_border_clip_rect );

DirectDraw->Log(LOG_OUT,"<<FORM>>::PaintForm()","");
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ExitButtonClick(TObject *Sender)
{
if (Enabled) Close();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ResetCountersClick(TObject *Sender)
{
fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::EnableDirectXButtonClick(TObject *Sender)
{
DirectDraw->Enabled = !DirectDraw->Enabled;
Repaint();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::UpdateDirectXMessages()
{
if (DirectDraw->ErrorStrings)
	{
	DirectDrawErrorMessagesBox->Items->Clear();
	for (int i=0;i<DirectDraw->ErrorStrings->Count;i++)
		DirectDrawErrorMessagesBox->Items->Add(DirectDraw->ErrorStrings->Strings[i]);
	DirectDraw->ErrorCount = 0;
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormCreate(TObject *Sender)
{
bool all_enabled = true;
DirectDraw->Enabled = true;
if (DirectDraw->Enabled)	
	{
	if (DirectDraw->SetDisplayMode(640,480,8))
    	{
	    AnimationTimer->Interval = AutoFlipTimer->Interval;
	    AnimationTimer->Enabled = true;
	    AutoFlipTimer->Enabled = true;
        all_enabled = true;
        }
    }
else
	ShowMessage("DirectDraw could not be enabled");

if (!all_enabled)
	{
    for (int i=0;i<ControlCount;i++) Controls[i]->Enabled = false;
    AnimationTimer->Enabled = false;
    AutoFlipTimer->Enabled = false;
    }

Left=0; Width=DirectDraw->Width-Left;
Top=5; Height=DirectDraw->Height-Top;
DirectDrawErrorMessagesBox->Height -= 5;

fFramesFlipped = 0; fSeconds = 0; fAnimationTicks = 0;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormActivate(TObject *Sender)
{
AnimationTimer->Enabled = true;
if (DirectDraw->ErrorValue == DD_OK)
	{
    DirectXInstalledLabel->Font->Color = clGreen;
    DirectXInstalledLabel->Caption = "DirectX is Installed";
    }
else
	{
    DirectXInstalledLabel->Font->Color = clRed;
    DirectXInstalledLabel->Caption = "DirectX is Not Installed";
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormDeactivate(TObject *Sender)
{
AnimationTimer->Enabled = false;
AutoFlipTimer->Enabled = false;
DirectDraw->Enabled = false;
DirectDrawErrorMessagesBox->Items->Clear();
}
//---------------------------------------------------------------------------
// yuk, temporary debugging vars, etc.
//---------------------------------------------------------------------------
TDateTime last_time = Time();
int PageNumber;
int x = 0;
int y = 0;
int xincr = 1;
int yincr = 1;
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::FormPaint(TObject *Sender)
{
// --------- debugging display ----------
Canvas->Pen->Color = clRed;
Canvas->Font->Color = clWhite;
Canvas->TextOut(200,0,"PAGE # " + IntToStr(PageNumber));

// --------- normal FormPaint() drawing code using the VCL or GDI with Canvas->Handle as the DC

// display the number of frames, etc. being drawn
TDateTime current_time = Time();
unsigned short chour,cmin,csec,cmsec;
unsigned short lhour,lmin,lsec,lmsec;
current_time.DecodeTime(&chour,&cmin,&csec,&cmsec);
last_time.DecodeTime(&lhour,&lmin,&lsec,&lmsec);
int seconds_elapsed = 0;
if (csec != lsec)
	{
    if (csec < lsec) seconds_elapsed += 60;
    seconds_elapsed += (csec - lsec);
    }
fSeconds += seconds_elapsed;
last_time = current_time;

++PageNumber %= 2;	// 0..1 only

TicksLabel->Caption = "Ticks: " + IntToStr(fAnimationTicks);
FramesLabel->Caption = "Frames: " + IntToStr(fFramesFlipped);
SecondsLabel->Caption = "Seconds: " + IntToStr(fSeconds);
if (fSeconds != 0)
	{
    FramesPerSecondLabel->Caption = "FPS: " + FloatToStrF((float)fFramesFlipped / (float)fSeconds, ffFixed, 6, 2);
	TicksPerSecondLabel->Caption = "TPS: " + FloatToStrF((float)fAnimationTicks / (float)fSeconds, ffFixed, 6, 2);
	}

// display a box moving about the screen (note: x,y,xincr,yincr global & just above function)
fColor++;
x += xincr;
y += yincr;

if (x<0) xincr = 1;
if (y<0) yincr = 1;
if (x+400>Width) xincr = -1;
if (y+200>Height) yincr = -1;

Canvas->Pen->Color = (TColor) RGB(255,255,255);
Canvas->Brush->Color = (TColor) RGB((x*128)%256,(x*16)%256,(x*16+16*fColor)%256) ;
Canvas->Brush->Style = bsSolid;
Canvas->Rectangle( x,y, x+400,y+200 );

// make sure the scroll box of directx messages is realtime.....
UpdateDirectXMessages();

// also display some DirectX stats, etc.
if (DirectDraw->Enabled)
	{
    DirectXEnabledLabel->Font->Color = clGreen;
    DirectXEnabledLabel->Caption = "DirectX drawing is enabled";
    }
else
	{
    DirectXEnabledLabel->Font->Color = clRed;
    DirectXEnabledLabel->Caption = "DirectX drawing is not enabled.";
    }
if (DirectDraw->Debugging)
	{
    DirectXDebuggingLabel->Font->Color = clGreen;
    DirectXDebuggingLabel->Caption = "DirectX debugging is enabled";
    }
else
	{
    DirectXDebuggingLabel->Font->Color = clRed;
    DirectXDebuggingLabel->Caption = "DirectX debugging is not enabled";
    }

Canvas->Pixels[random(Width)][random(Height)] = clWhite;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::RepaintFormButtonClick(TObject *Sender)
{
Refresh();
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::ClearMessagesButtonClick(TObject *Sender)
{
if (Enabled)
	{
    DirectDraw->ErrorStrings->Clear();
	Refresh();
    }
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::BitBtn3Click(TObject *Sender)
{
if (Enabled) DirectDraw->Debugging = !DirectDraw->Debugging;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::SecondsTimerTimer(TObject *Sender)
{
fSeconds++;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::AnimationTimerTimer(TObject *Sender)
{
fAnimationTicks++;
}
//---------------------------------------------------------------------------
void __fastcall TDirectDrawForm::AutoFlipTimerTimer(TObject *Sender)
{
fFramesFlipped++;
Refresh();
}
//---------------------------------------------------------------------------

void __fastcall TDirectDrawForm::FormDestroy(TObject *Sender)
{
DirectDrawErrorMessagesBox->Items->Clear();
}
//---------------------------------------------------------------------------

